"""Clickhouse connector."""

import logging
import re
from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, List, Optional, Tuple, Type

import sqlparse
from sqlalchemy import MetaData, text

from dbgpt.core.awel.flow import (
    TAGS_ORDER_HIGH,
    ResourceCategory,
    auto_register_resource,
)
from dbgpt.datasource.parameter import BaseDatasourceParameters
from dbgpt.datasource.rdbms.base import RDBMSConnector
from dbgpt.util.i18n_utils import _
from dbgpt_ext.datasource.schema import DBType

logger = logging.getLogger(__name__)


@auto_register_resource(
    label=_("Clickhouse datasource"),
    category=ResourceCategory.DATABASE,
    tags={"order": TAGS_ORDER_HIGH},
    description=_(
        "Columnar database for high-performance analytics and real-time queries."
    ),
)
@dataclass
class ClickhouseParameters(BaseDatasourceParameters):
    """Clickhouse connection parameters."""

    __type__ = "clickhouse"

    host: str = field(metadata={"help": _("Database host, e.g., localhost")})
    port: int = field(metadata={"help": _("Database port, e.g., 8123")})
    user: str = field(metadata={"help": _("Database user to connect")})
    database: str = field(metadata={"help": _("Database name")})
    engine: str = field(
        default="MergeTree", metadata={"help": _("Storage engine, e.g., MergeTree")}
    )
    password: str = field(
        default="${env:DBGPT_DB_PASSWORD}",
        metadata={
            "help": _(
                "Database password, you can write your password directly, of course, "
                "you can also use environment variables, such as "
                "${env:DBGPT_DB_PASSWORD}"
            ),
            "tags": "privacy",
        },
    )
    http_pool_maxsize: int = field(
        default=16, metadata={"help": _("http pool maxsize")}
    )
    http_pool_num_pools: int = field(
        default=12, metadata={"help": _("http pool num_pools")}
    )
    connect_timeout: int = field(
        default=15, metadata={"help": _("Database connect timeout, default 15s")}
    )
    distributed_ddl_task_timeout: int = field(
        default=300, metadata={"help": _("Distributed ddl task timeout, default 300s")}
    )

    def create_connector(self) -> "ClickhouseConnector":
        """Create clickhouse connector."""
        return ClickhouseConnector.from_parameters(self)

    def db_url(self, ssl=False, charset=None):
        raise NotImplementedError("Clickhouse does not support db_url")


class ClickhouseConnector(RDBMSConnector):
    """Clickhouse connector."""

    """db type"""
    db_type: str = "clickhouse"
    """db driver"""
    driver: str = "clickhouse"
    """db dialect"""
    db_dialect: str = "clickhouse"

    client: Any = None

    def __init__(self, client, engine, **kwargs):
        """Create a new ClickhouseConnector from client."""
        self.client = client

        self._all_tables = set()
        self.view_support = False
        self._usable_tables = set()
        self._include_tables = set()
        self._ignore_tables = set()
        self._custom_table_info = set()
        self._indexes_in_table_info = set()
        self._usable_tables = set()
        self._usable_tables = set()
        self._sample_rows_in_table_info = set()

        self._metadata = MetaData()

    @classmethod
    def param_class(cls) -> Type[ClickhouseParameters]:
        """Return the parameter class."""
        return ClickhouseParameters

    @classmethod
    def from_parameters(cls, parameters: ClickhouseParameters) -> "ClickhouseConnector":
        """Create a new ClickhouseConnector from parameters."""
        return cls.from_uri_db(
            parameters.host,
            parameters.port,
            parameters.user,
            parameters.password,
            parameters.database,
            parameters.http_pool_maxsize,
            parameters.http_pool_num_pools,
            parameters.connect_timeout,
            parameters.distributed_ddl_task_timeout,
            parameters.engine,
        )

    @classmethod
    def from_uri_db(
        cls,
        host: str,
        port: int,
        user: str,
        pwd: str,
        db_name: str,
        http_pool_maxsize: int = 16,
        http_pool_num_pools: int = 12,
        connect_timeout: int = 15,
        distributed_ddl_task_timeout: int = 300,
        engine: str = "MergeTree",
        **kwargs: Any,
    ) -> "ClickhouseConnector":
        """Create a new ClickhouseConnector from host, port, user, pwd, db_name."""
        import clickhouse_connect
        from clickhouse_connect.driver import httputil

        # Lazy import

        big_pool_mgr = httputil.get_pool_manager(
            maxsize=http_pool_maxsize, num_pools=http_pool_num_pools
        )
        client = clickhouse_connect.get_client(
            host=host,
            user=user,
            password=pwd,
            port=port,
            connect_timeout=connect_timeout,
            database=db_name,
            settings={"distributed_ddl_task_timeout": distributed_ddl_task_timeout},
            pool_mgr=big_pool_mgr,
        )

        cls.client = client
        return cls(client, engine=engine, **kwargs)

    def get_table_names(self):
        """Get all table names."""
        session = self.client

        with session.query_row_block_stream("SHOW TABLES") as stream:
            tables = [row[0] for block in stream for row in block]
            return tables

    def get_indexes(self, table_name: str) -> List[Dict]:
        """Get table indexes about specified table.

        Args:
            table_name (str): table name
        Returns:
            indexes: List[Dict], eg:[{'name': 'idx_key', 'column_names': ['id']}]
        """
        session = self.client

        _query_sql = f"""
                    SELECT name AS table, primary_key from system.tables where
                     database ='{self.client.database}' and table = '{table_name}'
                """
        with session.query_row_block_stream(_query_sql) as stream:
            indexes = [block for block in stream]  # noqa
            return [
                {"name": "primary_key", "column_names": column_names.split(",")}
                for table, column_names in indexes[0]
            ]

    @property
    def table_info(self) -> str:
        """Get table info."""
        return self.get_table_info()

    def get_table_info(self, table_names: Optional[List[str]] = None) -> str:
        """Get information about specified tables.

        Follows best practices as specified in: Rajkumar et al, 2022
        (https://arxiv.org/abs/2204.00498)

        If `sample_rows_in_table_info`, the specified number of sample rows will be
        appended to each table description. This can increase performance as
        demonstrated in the paper.
        """
        # TODO:
        return ""

    def get_show_create_table(self, table_name):
        """Get table show create table about specified table."""
        result = self.client.command(text(f"SHOW CREATE TABLE  {table_name}"))

        ans = result
        ans = re.sub(r"\s*ENGINE\s*=\s*MergeTree\s*", " ", ans, flags=re.IGNORECASE)
        ans = re.sub(
            r"\s*DEFAULT\s*CHARSET\s*=\s*\w+\s*", " ", ans, flags=re.IGNORECASE
        )
        ans = re.sub(r"\s*SETTINGS\s*\s*\w+\s*", " ", ans, flags=re.IGNORECASE)
        return ans

    def get_columns(self, table_name: str) -> List[Dict]:
        """Get columns.

        Args:
            table_name (str): str
        Returns:
            List[Dict], which contains name: str, type: str,
                default_expression: str, is_in_primary_key: bool, comment: str
                eg:[{'name': 'id', 'type': 'UInt64', 'default_expression': '',
                'is_in_primary_key': True, 'comment': 'id'}, ...]
        """
        fields = self.get_fields(table_name)
        return [
            {"name": name, "comment": comment, "type": column_type}
            for name, column_type, _, _, comment in fields[0]
        ]

    @property
    def dialect(self) -> str:
        """Return string representation of dialect to use."""
        return ""

    def get_fields(self, table_name, db_name=None) -> List[Tuple]:
        """Get column fields about specified table."""
        session = self.client
        _query_sql = f"""
            SELECT name, type, default_expression, is_in_primary_key, comment
                from system.columns where table='{table_name}'
        """.format(table_name)
        if db_name is not None:
            _query_sql += f" AND database='{db_name}'"
        with session.query_row_block_stream(_query_sql) as stream:
            fields = [block for block in stream]  # noqa
            return fields

    def get_users(self):
        """Get user info."""
        return []

    def get_grants(self):
        """Get grants."""
        return []

    def get_collation(self):
        """Get collation."""
        return "UTF-8"

    def get_charset(self):
        """Get character_set."""
        return "UTF-8"

    def get_database_names(self):
        """Get database names."""
        session = self.client

        with session.command("SHOW DATABASES") as stream:
            databases = [
                row[0]
                for block in stream
                for row in block
                if row[0]
                not in ("INFORMATION_SCHEMA", "system", "default", "information_schema")
            ]
            return databases

    def run(self, command: str, fetch: str = "all") -> List:
        """Execute sql command."""
        # TODO need to be implemented
        logger.info("SQL:" + command)
        if not command or len(command) < 0:
            return []
        _, ttype, sql_type, table_name = self.__sql_parse(command)
        if ttype == sqlparse.tokens.DML:
            if sql_type == "SELECT":
                return self._query(command, fetch)
            else:
                self._write(command)
                select_sql = self.convert_sql_write_to_select(command)
                logger.info(f"write result query:{select_sql}")
                return self._query(select_sql)
        else:
            logger.info(
                "DDL execution determines whether to enable through configuration "
            )

            cursor = self.client.command(command)

            if cursor.written_rows:
                result = cursor.result_rows
                field_names = result.column_names

                result = list(result)
                result.insert(0, field_names)
                logger.info("DDL Result:" + str(result))
                if not result:
                    # return self._query(f"SHOW COLUMNS FROM {table_name}")
                    return self.get_simple_fields(table_name)
                return result
            else:
                return self.get_simple_fields(table_name)

    def get_simple_fields(self, table_name):
        """Get column fields about specified table."""
        return self._query(f"SHOW COLUMNS FROM {table_name}")

    def get_current_db_name(self):
        """Get current database name."""
        return self.client.database

    def get_table_comments(self, db_name: str):
        """Get table comments."""
        session = self.client

        _query_sql = f"""
                SELECT table, comment FROM system.tables WHERE database = '{db_name}'
            """.format(db_name)

        with session.query_row_block_stream(_query_sql) as stream:
            table_comments = [row for block in stream for row in block]
            return table_comments

    def get_table_comment(self, table_name: str) -> Dict:
        """Get table comment.

        Args:
            table_name (str): table name
        Returns:
            comment: Dict, which contains text: Optional[str], eg:["text": "comment"]
        """
        session = self.client

        _query_sql = f"""
                SELECT table, comment FROM system.tables WHERE
                 database = '{self.client.database}'and table = '{table_name}'
                 """.format(self.client.database)

        with session.query_row_block_stream(_query_sql) as stream:
            table_comments = [row for block in stream for row in block]
            return [{"text": comment} for table_name, comment in table_comments][0]

    def get_column_comments(self, db_name, table_name):
        """Get column comments."""
        session = self.client
        _query_sql = f"""
            select name column, comment from  system.columns where database='{db_name}'
             and table='{table_name}'
        """.format(db_name, table_name)

        with session.query_row_block_stream(_query_sql) as stream:
            column_comments = [row for block in stream for row in block]
            return column_comments

    def table_simple_info(self):
        """Get table simple info."""
        # group_concat() not supported in clickhouse, use arrayStringConcat+groupArray
        # instead; and quotes need to be escaped

        _sql = f"""
            SELECT concat(TABLE_NAME, '(', arrayStringConcat(
                groupArray(column_name), '-'), ')') AS schema_info
            FROM INFORMATION_SCHEMA.COLUMNS
            WHERE table_schema = '{self.get_current_db_name()}'
            GROUP BY TABLE_NAME
        """
        with self.client.query_row_block_stream(_sql) as stream:
            return [row[0] for block in stream for row in block]

    def _write(self, write_sql: str):
        """Execute write sql.

        Args:
            write_sql (str): sql string
        """
        # TODO need to be implemented
        logger.info(f"Write[{write_sql}]")
        result = self.client.command(write_sql)
        logger.info(f"SQL[{write_sql}], result:{result.written_rows}")

    def _query(self, query: str, fetch: str = "all"):
        """Query data from clickhouse.

        Args:
            query (str): sql string
            fetch (str, optional): "one" or "all". Defaults to "all".

        Raises:
            ValueError: Error

        Returns:
            _type_: List<Result>
        """
        # TODO need to be implemented
        logger.info(f"Query[{query}]")

        if not query:
            return []

        cursor = self.client.query(query)
        if fetch == "all":
            result = cursor.result_rows
        elif fetch == "one":
            result = cursor.first_row
        else:
            raise ValueError("Fetch parameter must be either 'one' or 'all'")

        field_names = cursor.column_names
        result.insert(0, field_names)
        return result

    def __sql_parse(self, sql):
        sql = sql.strip()
        parsed = sqlparse.parse(sql)[0]
        sql_type = parsed.get_type()
        if sql_type == "CREATE":
            table_name = self._extract_table_name_from_ddl(parsed)
        else:
            table_name = parsed.get_name()

        first_token = parsed.token_first(skip_ws=True, skip_cm=False)
        ttype = first_token.ttype
        logger.info(
            f"SQL:{sql}, ttype:{ttype}, sql_type:{sql_type}, table:{table_name}"
        )
        return parsed, ttype, sql_type, table_name

    def _sync_tables_from_db(self) -> Iterable[str]:
        """Read table information from database."""
        # TODO Use a background thread to refresh periodically

        # SQL will raise error with schema
        _schema = (
            None if self.db_type == DBType.SQLite.value() else self._engine.url.database
        )
        # including view support by adding the views as well as tables to the all
        # tables list if view_support is True
        self._all_tables = set(
            self._inspector.get_table_names(schema=_schema)
            + (
                self._inspector.get_view_names(schema=_schema)
                if self.view_support
                else []
            )
        )
        return self._all_tables

    def close(self):
        """Close the connection.

        TODO: implement this method
        """
        pass



from unittest.mock import MagicMock, patch

import pytest

from ..docx import DocxKnowledge


@pytest.fixture
def mock_docx_document():
    mock_document = MagicMock()
    mock_document.paragraphs = [
        MagicMock(text="This is the first paragraph."),
        MagicMock(text="This is the second paragraph."),
    ]
    with patch("docx.Document", return_value=mock_document):
        yield mock_document


def test_load_from_docx(mock_docx_document):
    file_path = "test_document.docx"
    knowledge = DocxKnowledge(file_path=file_path)
    documents = knowledge._load()

    assert len(documents) == 1
    assert (
        documents[0].content
        == "This is the first paragraph.\nThis is the second paragraph."
    )
    assert documents[0].metadata["source"] == file_path



"""Define the CommunitySummaryKnowledgeGraph."""

import logging
import os
import uuid
from typing import List, Optional, Tuple

from dbgpt.core import Chunk, Embeddings, LLMClient
from dbgpt.core.awel.flow import Parameter, ResourceCategory, register_resource
from dbgpt.storage.graph_store.base import GraphStoreConfig
from dbgpt.storage.knowledge_graph.base import ParagraphChunk
from dbgpt.storage.vector_store.base import VectorStoreConfig
from dbgpt.storage.vector_store.filters import MetadataFilters
from dbgpt.util.i18n_utils import _
from dbgpt_ext.rag.retriever.graph_retriever.graph_retriever import GraphRetriever
from dbgpt_ext.rag.transformer.community_summarizer import CommunitySummarizer
from dbgpt_ext.rag.transformer.graph_embedder import GraphEmbedder
from dbgpt_ext.rag.transformer.graph_extractor import GraphExtractor
from dbgpt_ext.rag.transformer.text_embedder import TextEmbedder
from dbgpt_ext.storage.graph_store.tugraph_store import TuGraphStoreConfig
from dbgpt_ext.storage.knowledge_graph.community.community_store import CommunityStore
from dbgpt_ext.storage.knowledge_graph.knowledge_graph import (
    GRAPH_PARAMETERS,
    BuiltinKnowledgeGraph,
)

logger = logging.getLogger(__name__)


@register_resource(
    _("Community Summary KG Config"),
    "community_summary_kg_config",
    category=ResourceCategory.KNOWLEDGE_GRAPH,
    description=_("community Summary kg Config."),
    parameters=[
        *GRAPH_PARAMETERS,
        Parameter.build_from(
            _("Knowledge Graph Type"),
            "graph_store_type",
            str,
            description=_("graph store type."),
            optional=True,
            default="TuGraph",
        ),
        Parameter.build_from(
            _("LLM Client"),
            "llm_client",
            LLMClient,
            description=_("llm client for extract graph triplets."),
        ),
        Parameter.build_from(
            _("LLM Model Name"),
            "model_name",
            str,
            description=_("llm model name."),
            optional=True,
            default=None,
        ),
        Parameter.build_from(
            _("Vector Store Type"),
            "vector_store_type",
            str,
            description=_("vector store type."),
            optional=True,
            default="Chroma",
        ),
        Parameter.build_from(
            _("Topk of Knowledge Graph Extract"),
            "extract_topk",
            int,
            description=_("Topk of knowledge graph extract"),
            optional=True,
            default=5,
        ),
        Parameter.build_from(
            _("Recall Score of Knowledge Graph Extract"),
            "extract_score_threshold",
            float,
            description=_("Recall score of knowledge graph extract"),
            optional=True,
            default=0.3,
        ),
        Parameter.build_from(
            _("Recall Score of Community Search in Knowledge Graph"),
            "community_topk",
            int,
            description=_("Recall score of community search in knowledge graph"),
            optional=True,
            default=50,
        ),
        Parameter.build_from(
            _("Recall Score of Community Search in Knowledge Graph"),
            "community_score_threshold",
            float,
            description=_("Recall score of community search in knowledge graph"),
            optional=True,
            default=0.3,
        ),
        Parameter.build_from(
            _("Enable the graph search for documents and chunks"),
            "triplet_graph_enabled",
            bool,
            description=_("Enable the graph search for triplets"),
            optional=True,
            default=True,
        ),
        Parameter.build_from(
            _("Enable the graph search for documents and chunks"),
            "document_graph_enabled",
            bool,
            description=_("Enable the graph search for documents and chunks"),
            optional=True,
            default=True,
        ),
        Parameter.build_from(
            _("Top size of knowledge graph chunk search"),
            "knowledge_graph_chunk_search_top_size",
            int,
            description=_("Top size of knowledge graph chunk search"),
            optional=True,
            default=5,
        ),
        Parameter.build_from(
            _("Batch size of triplets extraction from the text"),
            "knowledge_graph_extraction_batch_size",
            int,
            description=_("Batch size of triplets extraction from the text"),
            optional=True,
            default=20,
        ),
        Parameter.build_from(
            _("Batch size of parallel community building process"),
            "community_summary_batch_size",
            int,
            description=_("TBatch size of parallel community building process"),
            optional=True,
            default=20,
        ),
    ],
)
@register_resource(
    _("Community Summary Knowledge Graph"),
    "community_summary_knowledge_graph",
    category=ResourceCategory.KNOWLEDGE_GRAPH,
    description=_("Community Summary Knowledge Graph."),
    parameters=[
        Parameter.build_from(
            _("Graph Store Config"),
            "config",
            GraphStoreConfig,
            description=_("graph store config."),
        ),
        Parameter.build_from(
            _("Graph Store Name"),
            "name",
            str,
            optional=True,
            default="dbgpt",
            description=_("Graph Store Name"),
        ),
        Parameter.build_from(
            _("LLM Client"),
            "llm_client",
            LLMClient,
            description=_("llm client for extract graph triplets."),
        ),
        Parameter.build_from(
            _("LLM Model Name"),
            "llm_model",
            str,
            description=_("kg extract llm model name."),
            optional=True,
            default=None,
        ),
    ],
)
class CommunitySummaryKnowledgeGraph(BuiltinKnowledgeGraph):
    """Community summary knowledge graph class."""

    def __init__(
        self,
        config: GraphStoreConfig,
        name: Optional[str] = "dbgpt",
        llm_client: Optional[LLMClient] = None,
        llm_model: Optional[str] = None,
        kg_extract_top_k: Optional[int] = 5,
        kg_extract_score_threshold: Optional[float] = 0.3,
        kg_community_top_k: Optional[int] = 50,
        kg_community_score_threshold: Optional[float] = 0.3,
        kg_triplet_graph_enabled: Optional[bool] = True,
        kg_document_graph_enabled: Optional[bool] = True,
        kg_chunk_search_top_k: Optional[int] = 5,
        kg_extraction_batch_size: Optional[int] = 3,
        kg_community_summary_batch_size: Optional[int] = 20,
        kg_embedding_batch_size: Optional[int] = 20,
        kg_similarity_top_k: Optional[int] = 5,
        kg_similarity_score_threshold: Optional[float] = 0.7,
        kg_enable_text_search: Optional[float] = False,
        kg_text2gql_model_enabled: Optional[bool] = False,
        kg_text2gql_model_name: Optional[str] = None,
        embedding_fn: Optional[Embeddings] = None,
        vector_store_config: Optional["VectorStoreConfig"] = None,
        kg_max_chunks_once_load: Optional[int] = 10,
        kg_max_threads: Optional[int] = 1,
    ):
        """Initialize community summary knowledge graph class."""
        super().__init__(
            config=config, name=name, llm_client=llm_client, llm_model=llm_model
        )
        self._config = config

        self._vector_store_type = config.get_type_value() or os.getenv(
            "VECTOR_STORE_TYPE"
        )
        self._extract_topk = int(
            kg_extract_top_k or os.getenv("KNOWLEDGE_GRAPH_EXTRACT_SEARCH_TOP_SIZE")
        )
        self._extract_score_threshold = float(
            kg_extract_score_threshold
            or os.getenv("KNOWLEDGE_GRAPH_EXTRACT_SEARCH_RECALL_SCORE")
        )
        self._community_topk = int(
            kg_community_top_k or os.getenv("KNOWLEDGE_GRAPH_COMMUNITY_SEARCH_TOP_SIZE")
        )
        self._community_score_threshold = float(
            kg_community_score_threshold
            or os.getenv("KNOWLEDGE_GRAPH_COMMUNITY_SEARCH_RECALL_SCORE")
        )
        self._document_graph_enabled = kg_document_graph_enabled or (
            os.getenv("DOCUMENT_GRAPH_ENABLED", "").lower() == "true"
        )
        self._triplet_graph_enabled = kg_triplet_graph_enabled or (
            os.getenv("TRIPLET_GRAPH_ENABLED", "").lower() == "true"
        )
        self._triplet_extraction_batch_size = int(
            kg_extraction_batch_size
            or os.getenv("KNOWLEDGE_GRAPH_EXTRACTION_BATCH_SIZE")
        )
        self._triplet_embedding_batch_size = int(
            kg_embedding_batch_size or os.getenv("KNOWLEDGE_GRAPH_EMBEDDING_BATCH_SIZE")
        )
        self._community_summary_batch_size = int(
            kg_community_summary_batch_size or os.getenv("COMMUNITY_SUMMARY_BATCH_SIZE")
        )
        self._embedding_fn = embedding_fn
        self._vector_store_config = vector_store_config

        self._graph_extractor = GraphExtractor(
            self._llm_client,
            self._model_name,
            vector_store_config.create_store(
                name=name + "_CHUNK_HISTORY", embedding_fn=embedding_fn
            ),
            index_name=name,
            max_chunks_once_load=kg_max_chunks_once_load,
            max_threads=kg_max_threads,
            top_k=kg_extract_top_k,
            score_threshold=kg_extract_score_threshold,
        )

        self._graph_embedder = GraphEmbedder(embedding_fn)
        self._text_embedder = TextEmbedder(embedding_fn)

        # def community_store_configure(name: str, cfg: VectorStoreConfig):
        #     cfg.name = name
        #     cfg.embedding_fn = self._embedding_fn
        #     cfg.max_chunks_once_load = max_chunks_once_load
        #     cfg.max_threads = max_threads
        #     cfg.user = config.user
        #     cfg.password = config.password
        #     cfg.topk = self._community_topk
        #     cfg.score_threshold = self._community_score_threshold

        self._community_store = CommunityStore(
            self._graph_store_adapter,
            CommunitySummarizer(self._llm_client, self._model_name),
            vector_store_config.create_store(
                name=name + "_COMMUNITY_SUMMARY", embedding_fn=embedding_fn
            ),
            index_name=name,
            max_chunks_once_load=kg_max_chunks_once_load,
            max_threads=kg_max_threads,
            top_k=kg_community_top_k,
            score_threshold=kg_extract_score_threshold,
        )

        self._graph_retriever = GraphRetriever(
            self._graph_store_adapter,
            llm_client=llm_client,
            llm_model=llm_model,
            triplet_graph_enabled=kg_triplet_graph_enabled,
            document_graph_enabled=kg_document_graph_enabled,
            extract_top_k=kg_extract_top_k,
            kg_chunk_search_top_k=kg_chunk_search_top_k,
            similarity_top_k=kg_similarity_top_k,
            similarity_score_threshold=kg_similarity_score_threshold,
            embedding_fn=embedding_fn,
            embedding_batch_size=kg_embedding_batch_size,
            text2gql_model_enabled=kg_text2gql_model_enabled,
            text2gql_model_name=kg_text2gql_model_name,
        )

    def get_config(self) -> TuGraphStoreConfig:
        """Get the knowledge graph config."""
        return self._config

    @property
    def embeddings(self) -> Embeddings:
        """Get the knowledge graph config."""
        return self._embedding_fn

    async def aload_document(
        self, chunks: List[Chunk], file_id: Optional[str] = None
    ) -> List[str]:
        """Extract and persist graph from the document file."""
        if not self.vector_name_exists():
            self._graph_store_adapter.create_graph(self._graph_name)
        await self._aload_document_graph(chunks)
        await self._aload_triplet_graph(chunks, file_id)
        await self._community_store.build_communities(
            batch_size=self._community_summary_batch_size
        )

        return [chunk.chunk_id for chunk in chunks]

    async def _aload_document_graph(self, chunks: List[Chunk]) -> None:
        """Load the knowledge graph from the chunks.

        The chunks include the doc structure.
        """
        if not self._document_graph_enabled:
            return

        _chunks: List[ParagraphChunk] = [
            ParagraphChunk.model_validate(chunk.model_dump()) for chunk in chunks
        ]
        documment_chunk, paragraph_chunks = self._load_chunks(_chunks)

        if self._graph_store.enable_similarity_search:
            # Add embeddings from chunk content
            texts: List[str] = [chunk.content for chunk in paragraph_chunks]

            embeddings = await self._text_embedder.batch_embed(
                inputs=texts,
                batch_size=self._triplet_embedding_batch_size,
            )

            for idx, chunk in enumerate(paragraph_chunks):
                chunk.embedding = embeddings[idx]

        # upsert the document and chunks vertices
        self._graph_store_adapter.upsert_documents(iter([documment_chunk]))
        self._graph_store_adapter.upsert_chunks(iter(paragraph_chunks))

        # upsert the document structure
        for chunk_index, chunk in enumerate(paragraph_chunks):
            # document -> include -> chunk
            if chunk.parent_is_document:
                self._graph_store_adapter.upsert_doc_include_chunk(chunk=chunk)
            else:  # chunk -> include -> chunk
                self._graph_store_adapter.upsert_chunk_include_chunk(chunk=chunk)

            # chunk -> next -> chunk
            if chunk_index >= 1:
                self._graph_store_adapter.upsert_chunk_next_chunk(
                    chunk=paragraph_chunks[chunk_index - 1], next_chunk=chunk
                )

    async def _aload_triplet_graph(
        self, chunks: List[Chunk], file_id: Optional[str] = None
    ) -> None:
        """Load the knowledge graph from the chunks.

        The chunks include the doc structure.
        """
        if not self._triplet_graph_enabled:
            return

        document_graph_enabled = self._document_graph_enabled

        # Extract the triplets from the chunks, and return the list of graphs
        # in the same order as the input texts
        graphs_list = await self._graph_extractor.batch_extract(
            [chunk.content for chunk in chunks],
            batch_size=self._triplet_extraction_batch_size,
            file_id=file_id,
        )
        if not graphs_list:
            raise ValueError("No graphs extracted from the chunks")

        # If enable the similarity search, add the embedding to the graphs
        if self._graph_store.enable_similarity_search:
            for idx, graphs in enumerate(graphs_list):
                embeded_graphs = await self._graph_embedder.batch_embed(
                    inputs=graphs,
                    batch_size=self._triplet_embedding_batch_size,
                )
                graphs_list[idx] = embeded_graphs

        # Upsert the graphs into the graph store
        for idx, graphs in enumerate(graphs_list):
            for graph in graphs:
                if document_graph_enabled:
                    # Append the chunk id to the edge
                    for edge in graph.edges():
                        edge.set_prop("_chunk_id", chunks[idx].chunk_id)
                        graph.append_edge(edge=edge)

                # Upsert the graph
                self._graph_store_adapter.upsert_graph(graph)

                # chunk -> include -> entity
                if document_graph_enabled:
                    for vertex in graph.vertices():
                        self._graph_store_adapter.upsert_chunk_include_entity(
                            chunk=chunks[idx], entity=vertex
                        )

    def _load_chunks(
        self, chunks: List[ParagraphChunk]
    ) -> Tuple[ParagraphChunk, List[ParagraphChunk]]:
        """Load the chunks, and add the parent-child relationship within chunks."""
        # init default document
        doc_id = str(uuid.uuid4())
        doc_name = os.path.basename(chunks[0].metadata["source"] or "Text_Node")
        doc_chunk = ParagraphChunk(
            chunk_id=doc_id,
            chunk_name=doc_name,
        )

        # chunk.metadata = {"Header0": "title",
        # "Header1": "title", ..., "source": "source_path"}  # noqa: E501
        for chunk_index, chunk in enumerate(chunks):
            parent = None
            directory_keys = list(chunk.metadata.keys())[
                :-1
            ]  # ex: ['Header0', 'Header1', 'Header2', ...]
            parent_level = directory_keys[-2] if len(directory_keys) > 1 else None
            current_level = directory_keys[-1] if directory_keys else "Header0"

            chunk.chunk_name = chunk.metadata.get(current_level, "none_header_chunk")

            # Find the parent chunk for every chunk
            # parent chunk -> chunk
            if parent_level:
                for parent_direct in reversed(directory_keys[:-1]):
                    parent_titile = chunk.metadata.get(parent_direct, None)
                    for n in reversed(range(chunk_index)):
                        metadata = chunks[n].metadata
                        keys = list(metadata.keys())[:-1]
                        if (
                            metadata
                            and parent_direct == keys[-1]
                            and parent_titile == metadata.get(parent_direct)
                        ):
                            parent = chunks[n]
                            chunk.chunk_parent_id = parent.chunk_id
                            chunk.chunk_parent_name = parent_titile
                            chunk.parent_content = parent.content
                            break
                        if chunk_index - n > len(directory_keys):
                            break
                    if chunk.chunk_parent_id:
                        break

            if not chunk.chunk_parent_id:
                chunk.chunk_parent_id = doc_id
                chunk.chunk_parent_name = doc_name
                chunk.parent_content = ""
                chunk.parent_is_document = True

        return doc_chunk, chunks

    async def asimilar_search_with_scores(
        self,
        text,
        topk,
        score_threshold: float,
        filters: Optional[MetadataFilters] = None,
    ) -> List[Chunk]:
        """Retrieve relevant community summaries."""
        # Global search: retrieve relevant community summaries
        communities = await self._community_store.search_communities(text)
        summaries = [
            f"Section {i + 1}:\n{community.summary}"
            for i, community in enumerate(communities)
        ]
        context = "\n".join(summaries) if summaries else ""

        (
            subgraph,
            (
                subgraph_for_doc,
                text2gql_query,
            ),
        ) = await self._graph_retriever.retrieve(text)

        knowledge_graph_str = subgraph.format() if subgraph else ""
        knowledge_graph_for_doc_str = (
            subgraph_for_doc.format() if subgraph_for_doc else ""
        )
        if not (summaries or knowledge_graph_str or knowledge_graph_for_doc_str):
            return []

        # merge search results into context
        content = HYBRID_SEARCH_PT.format(
            context=context,
            query=text2gql_query,
            knowledge_graph=knowledge_graph_str,
            knowledge_graph_for_doc=knowledge_graph_for_doc_str,
        )
        logger.info(f"Final GraphRAG queried prompt:\n{content}")
        return [Chunk(content=content)]

    def truncate(self) -> List[str]:
        """Truncate knowledge graph."""
        logger.info("Truncate community store")
        self._community_store.truncate()
        logger.info("Truncate keyword extractor")
        self._keyword_extractor.truncate()
        logger.info("Truncate triplet extractor")
        self._graph_extractor.truncate()
        logger.info("Truncate graph embedder")
        self._graph_embedder.truncate()
        logger.info("Truncate text embedder")
        self._text_embedder.truncate()
        return [self._graph_name]

    def delete_vector_name(self, index_name: str):
        """Delete knowledge graph."""
        logger.info("Drop community store")
        self._community_store.drop()

        logger.info("Drop keyword extractor")
        self._keyword_extractor.drop()

        logger.info("Drop triplet extractor")
        self._graph_extractor.drop()

        logger.info("Drop graph embedder")
        self._graph_embedder.drop()

        logger.info("Drop text embedder")
        self._text_embedder.drop()


HYBRID_SEARCH_PT = """
=====
The following information from [Context], [Graph Query Statement], [Knowledge Graph], and [Original Text From RAG] can help you answer user questions better.

[Context]:
{context}

[Graph Query Statement]:
{query}

[Knowledge Graph]:
{knowledge_graph}

[Original Text From RAG]
{knowledge_graph_for_doc}
=====

You are very good at combining the [Context] information provided by the prompt word template with the [Knowledge Graph] information,
answering the user's questions accurately and appropriately, and ensuring that no information irrelevant to the context and knowledge graph is output.

## Role: GraphRAG Assistant

### Core Capabilities
0. Make sure DO NOT answer irrelevant questions from the user.

1. Information Processing
- Process contextual information across multiple sections ([Section] markers)
- Interpret knowledge graph relationships ((entity)-[relationship]->(entity))
- Synthesize information from both structured and unstructured sources

2. Response Generation
- Provide nuanced, multi-perspective answers
- Balance technical accuracy with conversational engagement
- Connect related concepts across different information sources
- Highlight uncertainties and limitations when appropriate

3. Interaction Style
- Maintain a natural, engaging conversation flow
- Ask clarifying questions when needed
- Provide examples and analogies to illustrate complex points
- Adapt explanation depth based on user's apparent expertise

4. Knowledge Integration
- Seamlessly blend information from:
  * Context sections
  * Knowledge graph relationships
  * Background knowledge (when appropriate)
- Prioritize relevance over comprehensiveness
- Acknowledge information gaps explicitly

5. Quality Assurance
- Verify logical consistency across sources
- Cross-reference relationships for validation
- Flag potential contradictions or ambiguities
- Provide confidence levels when appropriate

### Information Sources Handling
1. Context Processing [Context]
- Parse information from numbered sections systematically
- Identify key concepts and relationships within each section
- Track section dependencies and cross-references
- Prioritize recent/relevant sections for the query

2. Knowledge Graph Integration [Knowledge Graph]
- Parse Entities and Relationships sections separately
- Map entity-relationship-entity triples accurately
- Understand relationship directionality
- Use graph structure to find connected information

3. Original Text Reference [Original Text From RAG]
- The GraphRAG document directory is stored as an edge in relationships to show the hierarchy of the current source text in the entire document.
- Use as authoritative source for detailed information
- Cross-reference with Context and Knowledge Graph
- Extract supporting evidence and examples
- Resolve conflicts between sources using this as primary reference

4. Original Graph Query [Graph Query Statement]
- The graph query statement used if text2gql translation is successful
- Graph query will be empty if the translation failed
- Use the markdown code block format to highlight the graph query statement if the statement is not empty

### Output Format
1. Answer Structure
- Lead with a markdown code block to highlight the original cypher query statement from [Graph Query Statement] if it's not empty
- Demonstate synthesized core information
- Support with specific references to sources
- Include relevant entity-relationship pairs
- Conclude with confidence assessment
- Use the markdown format of the "quote" to highlight the original text (in details) from "GraphRAG"

=====
"""  # noqa: E501



"""GraphEmbedder class."""

import asyncio
import logging
from typing import List

from dbgpt.rag.transformer.base import EmbedderBase
from dbgpt.storage.graph_store.graph import Graph, GraphElemType

logger = logging.getLogger(__name__)


class GraphEmbedder(EmbedderBase):
    """GraphEmbedder class."""

    async def batch_embed(
        self,
        inputs: List[Graph],
        batch_size: int = 1,
    ) -> List[Graph]:
        """Embed graph from graphs in batches."""
        for graph in inputs:
            texts = []
            vectors = []

            # Get the text from graph
            for vertex in graph.vertices():
                if vertex.get_prop("vertex_type") == GraphElemType.CHUNK.value:
                    texts.append(vertex.get_prop("content"))
                elif vertex.get_prop("vertex_type") == GraphElemType.ENTITY.value:
                    texts.append(vertex.vid)
                else:
                    texts.append(" ")

            n_texts = len(texts)

            # Batch embedding
            for batch_idx in range(0, n_texts, batch_size):
                start_idx = batch_idx
                end_idx = min(start_idx + batch_size, n_texts)
                batch_texts = texts[start_idx:end_idx]

                # Create tasks
                embedding_tasks = [(self.embed(text)) for text in batch_texts]

                # Process embedding in parallel
                batch_results = await asyncio.gather(
                    *(task for task in embedding_tasks), return_exceptions=True
                )

                # Place results in the correct positions
                for idx, vector in enumerate(batch_results):
                    if isinstance(vector, Exception):
                        raise RuntimeError(f"Failed to embed text{idx}")
                    vectors.append(vector)

            # Push vectors back into Graph
            for vertex, vector in zip(graph.vertices(), vectors):
                vertex.set_prop("_embedding", vector)

        return inputs

    def truncate(self):
        """Do nothing by default."""

    def drop(self):
        """Do nothing by default."""



"""Dcoument Based Graph Retriever."""

import logging
from typing import List, Tuple, Union

from dbgpt.storage.graph_store.graph import Graph
from dbgpt_ext.rag.retriever.graph_retriever.base import GraphRetrieverBase

logger = logging.getLogger(__name__)


class DocumentGraphRetriever(GraphRetrieverBase):
    """Document Graph retriever class."""

    def __init__(
        self,
        graph_store_adapter,
        document_topk,
        similarity_search_topk,
        similarity_search_score_threshold,
    ):
        """Initialize Document Graph Retriever."""
        self._graph_store_adapter = graph_store_adapter
        self._document_topk = document_topk
        self._similarity_search_topk = similarity_search_topk
        self._similarity_search_score_threshold = similarity_search_score_threshold

    async def retrieve(
        self, input: Union[Graph, List[str], List[List[float]]]
    ) -> Tuple[Graph, None]:
        """Retrieve from document graph."""
        # If retrieve subgraph from triplet graph successfully
        if isinstance(input, Graph):
            # Get entities' vids from triplet subgraph
            keywords_for_document_graph = []
            for vertex in input.vertices():
                keywords_for_document_graph.append(vertex.name)
            # Using the vids to search chunks and doc
            # entities -> chunks -> doc
            subgraph_for_doc = self._graph_store_adapter.explore_docgraph_with_entities(
                subs=keywords_for_document_graph,
                topk=self._similarity_search_topk,
                score_threshold=self._similarity_search_score_threshold,
                limit=self._document_topk,
            )
        else:
            # Using subs to search chunks
            # subs -> chunks -> doc
            subgraph_for_doc = (
                self._graph_store_adapter.explore_docgraph_without_entities(
                    subs=input,
                    topk=self._similarity_search_topk,
                    score_threshold=self._similarity_search_score_threshold,
                    limit=self._document_topk,
                )
            )

        return subgraph_for_doc



from dataclasses import dataclass, field
from typing import Optional

from dbgpt.core.interface.file import StorageBackend, StorageBackendConfig
from dbgpt.util.i18n_utils import _


@dataclass
class OSSStorageConfig(StorageBackendConfig):
    __type__ = "oss"
    endpoint: str = field(
        metadata={
            "help": _(
                "The endpoint of the OSS server. "
                "e.g. https://oss-cn-hangzhou.aliyuncs.com"
            )
        },
    )
    region: str = field(
        metadata={"help": _("The region of the OSS server. e.g. cn-hangzhou")},
    )
    access_key_id: Optional[str] = field(
        default=None,
        metadata={
            "help": _(
                "The access key ID of the OSS server. You can also set it in the "
                "environment variable OSS_ACCESS_KEY_ID"
            ),
            "tags": "privacy",
        },
    )
    access_key_secret: Optional[str] = field(
        default=None,
        metadata={
            "help": _(
                "The access key secret of the OSS server. You can also set it in the "
                "environment variable OSS_ACCESS_KEY_SECRET"
            ),
            "tags": "privacy",
        },
    )
    use_environment_credentials: Optional[bool] = field(
        default=False,
        metadata={
            "help": _(
                "Whether to use the environment variables OSS_ACCESS_KEY_ID and "
                "OSS_ACCESS_KEY_SECRET as the credentials. Default is False."
            ),
        },
    )
    fixed_bucket: Optional[str] = field(
        default=None,
        metadata={
            "help": _(
                "The fixed bucket name to use. If set, all logical buckets in DB-GPT "
                "will be mapped to this bucket. We suggest you set this value to avoid "
                "bucket name conflicts."
            )
        },
    )
    bucket_prefix: Optional[str] = field(
        default="dbgpt-fs-",
        metadata={
            "help": _(
                "The prefix of the bucket name. If set, all logical buckets in DB-GPT "
                "will be prefixed with this value. Just work when fixed_bucket is None."
            )
        },
    )
    auto_create_bucket: Optional[bool] = field(
        default=True,
        metadata={
            "help": _(
                "Whether to create the bucket automatically if it does not exist. "
                "If set to False, the bucket must exist before using it."
            )
        },
    )
    save_chunk_size: Optional[int] = field(
        default=1024 * 1024,
        metadata={
            "help": _(
                "The chunk size when saving the file. When the file is larger 10x than "
                "this value, it will be uploaded in multiple parts. Default is 1M."
            )
        },
    )

    def create_storage(self) -> StorageBackend:
        from .oss_storage import AliyunOSSStorage

        return AliyunOSSStorage(
            endpoint=self.endpoint,
            region=self.region,
            access_key_id=self.access_key_id,
            access_key_secret=self.access_key_secret,
            use_environment_credentials=self.use_environment_credentials,
            fixed_bucket=self.fixed_bucket,
            bucket_prefix=self.bucket_prefix,
            auto_create_bucket=self.auto_create_bucket,
            save_chunk_size=self.save_chunk_size,
        )



"""Module for RAG operators."""

from dbgpt.rag.operators.chunk_manager import ChunkManagerOperator  # noqa: F401
from dbgpt.rag.operators.evaluation import RetrieverEvaluatorOperator  # noqa: F401
from dbgpt.rag.operators.rerank import RerankOperator  # noqa: F401
from dbgpt.rag.operators.rewrite import QueryRewriteOperator  # noqa: F401

from .db_schema import DBSchemaAssemblerOperator
from .embedding import (  # noqa: F401
    EmbeddingAssemblerOperator,
    EmbeddingRetrieverOperator,
)
from .full_text import FullTextStorageOperator  # noqa: F401
from .knowledge import ChunksToStringOperator, KnowledgeOperator  # noqa: F401
from .knowledge_graph import KnowledgeGraphOperator  # noqa: F401
from .process_branch import (
    KnowledgeProcessBranchOperator,  # noqa: F401
    KnowledgeProcessJoinOperator,
)
from .summary import SummaryAssemblerOperator  # noqa: F401
from .vector_store import VectorStorageOperator  # noqa: F401

__all__ = [
    "DBSchemaAssemblerOperator",
    "EmbeddingRetrieverOperator",
    "EmbeddingAssemblerOperator",
    "FullTextStorageOperator",
    "KnowledgeOperator",
    "KnowledgeGraphOperator",
    "KnowledgeProcessBranchOperator",
    "KnowledgeProcessJoinOperator",
    "ChunksToStringOperator",
    "SummaryAssemblerOperator",
    "VectorStorageOperator",
]



"""GraphExtractor class."""

import asyncio
import logging
import re
from typing import Dict, List, Optional

from dbgpt.core import Chunk, LLMClient
from dbgpt.rag.transformer.llm_extractor import LLMExtractor
from dbgpt.storage.graph_store.graph import Edge, Graph, MemoryGraph, Vertex
from dbgpt.storage.vector_store.base import VectorStoreBase

logger = logging.getLogger(__name__)


class GraphExtractor(LLMExtractor):
    """GraphExtractor class."""

    def __init__(
        self,
        llm_client: LLMClient,
        model_name: str,
        chunk_history: VectorStoreBase,
        index_name: str,
        max_chunks_once_load: Optional[int] = 10,
        max_threads: Optional[int] = 1,
        top_k: Optional[int] = 5,
        score_threshold: Optional[float] = 0.7,
    ):
        """Initialize the GraphExtractor."""
        super().__init__(llm_client, model_name, GRAPH_EXTRACT_PT_CN)
        self._chunk_history = chunk_history

        # config = self._chunk_history.get_config()

        self._vector_space = index_name
        self._max_chunks_once_load = max_chunks_once_load
        self._max_threads = max_threads
        self._topk = top_k
        self._score_threshold = score_threshold

    async def aload_chunk_context(
        self, texts: List[str], file_id: Optional[str] = None
    ) -> Dict[str, str]:
        """Load chunk context."""
        text_context_map: Dict[str, str] = {}

        for text in texts:
            # Load similar chunks
            chunks = await self._chunk_history.asimilar_search_with_scores(
                text, self._topk, self._score_threshold
            )
            history = [
                f"Section {i + 1}:\n{chunk.content}" for i, chunk in enumerate(chunks)
            ]

            # Save chunk to history
            # here we save the file_id into the metadata
            await self._chunk_history.aload_document_with_limit(
                [
                    Chunk(
                        content=text,
                        metadata={"relevant_cnt": len(history), "file_id": file_id},
                    )
                ],
                self._max_chunks_once_load,
                self._max_threads,
            )

            # Save chunk context to map
            context = "\n".join(history) if history else ""
            text_context_map[text] = context
        return text_context_map

    async def extract(self, text: str, limit: Optional[int] = None) -> List:
        """Extract graphs from text.

        Suggestion: to extract triplets in batches, call `batch_extract`.
        """
        # Load similar chunks
        text_context_map = await self.aload_chunk_context([text])
        context = text_context_map[text]

        # Extract with chunk history
        return await super()._extract(text, context, limit)

    async def batch_extract(
        self,
        texts: List[str],
        batch_size: int = 1,
        limit: Optional[int] = None,
        file_id: Optional[str] = None,
    ) -> Optional[List[List[Graph]]]:
        """Extract graphs from chunks in batches.

        Returns list of graphs in same order as input texts (text <-> graphs).
        """
        if batch_size < 1:
            raise ValueError("batch_size >= 1")

        # 1. Load chunk context
        text_context_map = await self.aload_chunk_context(texts, file_id)

        # Pre-allocate results list to maintain order
        graphs_list: List[List[Graph]] = [None] * len(texts)

        n_texts = len(texts)

        for batch_idx in range(0, n_texts, batch_size):
            start_idx = batch_idx
            end_idx = min(start_idx + batch_size, n_texts)
            batch_texts = texts[start_idx:end_idx]

            # 2. Create tasks with their original indices
            extraction_tasks = [
                (
                    idx,
                    self._extract(text, text_context_map[text], limit),
                )
                for idx, text in enumerate(batch_texts, start=start_idx)
            ]

            # 3. Process extraction in parallel while keeping track of indices
            batch_results = await asyncio.gather(
                *(task for _, task in extraction_tasks), return_exceptions=True
            )

            # 4. Place results in the correct positions
            for (idx, _), graphs in zip(extraction_tasks, batch_results):
                if isinstance(graphs, Exception):
                    raise RuntimeError(f"Failed to extract graph: {graphs}")
                if not isinstance(graphs, list) or not all(
                    isinstance(g, Graph) for g in graphs
                ):
                    raise RuntimeError(f"Invalid graph extraction result: {graphs}")
                graphs_list[idx] = graphs

        assert all(x is not None for x in graphs_list), "All positions should be filled"
        return graphs_list

    def _parse_response(self, text: str, limit: Optional[int] = None) -> List[Graph]:
        graph = MemoryGraph()
        edge_count = 0
        current_section = None
        for line in text.split("\n"):
            line = line.strip()
            if line in ["Entities:", "Relationships:"]:
                current_section = line[:-1]
            elif line and current_section:
                if current_section == "Entities":
                    match = re.match(r"\((.*?)#(.*?)\)", line)
                    if match:
                        name, summary = [part.strip() for part in match.groups()]
                        graph.upsert_vertex(
                            Vertex(name, description=summary, vertex_type="entity")
                        )
                elif current_section == "Relationships":
                    match = re.match(r"\((.*?)#(.*?)#(.*?)#(.*?)\)", line)
                    if match:
                        source, name, target, summary = [
                            part.strip() for part in match.groups()
                        ]
                        edge_count += 1
                        graph.append_edge(
                            Edge(
                                source,
                                target,
                                name,
                                description=summary,
                                edge_type="relation",
                            )
                        )

            if limit and edge_count >= limit:
                break

        return [graph]

    def truncate(self):
        """Truncate chunk history."""
        self._chunk_history.truncate()

    def drop(self):
        """Drop chunk history."""
        self._chunk_history.delete_vector_name(self._vector_space)


GRAPH_EXTRACT_PT_CN = (
    "## 角色\n"
    "你是一个知识图谱工程专家，非常擅长从文本中精确抽取知识图谱的实体"
    "（主体、客体）和关系，并能对实体和关系的含义做出恰当的总结性描述。\n"
    "\n"
    "## 技能\n"
    "### 技能 1: 实体抽取\n"
    "--请按照如下步骤抽取实体--\n"
    "1. 准确地识别文本中的实体信息，一般是名词、代词等。\n"
    "2. 准确地识别实体的修饰性描述，一般作为定语对实体特征做补充。\n"
    "3. 对相同概念的实体（同义词、别称、代指），请合并为单一简洁的实体名，"
    "并合并它们的描述信息。\n"
    "4. 对合并后的实体描述信息做简洁、恰当、连贯的总结。\n"
    "\n"
    "### 技能 2: 关系抽取\n"
    "--请按照如下步骤抽取关系--\n"
    "1. 准确地识别文本中实体之间的关联信息，一般是动词、代词等。\n"
    "2. 准确地识别关系的修饰性描述，一般作为状语对关系特征做补充。\n"
    "3. 对相同概念的关系（同义词、别称、代指），请合并为单一简洁的关系名，"
    "并合并它们的描述信息。\n"
    "4. 对合并后的关系描述信息做简洁、恰当、连贯的总结。\n"
    "\n"
    "### 技能 3: 关联上下文\n"
    "- 关联上下文来自与当前待抽取文本相关的前置段落内容，"
    "可以为知识抽取提供信息补充。\n"
    "- 合理利用提供的上下文信息，知识抽取过程中出现的内容引用可能来自关联上下文。\n"
    "- 不要对关联上下文的内容做知识抽取，而仅作为关联信息参考。\n"
    "- 关联上下文是可选信息，可能为空。\n"
    "\n"
    "## 约束条件\n"
    "- 如果文本已提供了图结构格式的数据，直接转换为输出格式返回，"
    "不要修改实体或ID名称。"
    "- 尽可能多的生成文本中提及的实体和关系信息，但不要随意创造不存在的实体和关系。\n"
    "- 确保以第三人称书写，从客观角度描述实体名称、关系名称，以及他们的总结性描述。\n"
    "- 尽可能多地使用关联上下文中的信息丰富实体和关系的内容，这非常重要。\n"
    "- 如果实体或关系的总结描述为空，不提供总结描述信息，不要生成无关的描述信息。\n"
    "- 如果提供的描述信息相互矛盾，请解决矛盾并提供一个单一、连贯的描述。\n"
    "- 实体和关系的名称或者描述文本出现#和:字符时，使用_字符替换，其他字符不要修改。"
    "- 避免使用停用词和过于常见的词汇。\n"
    "\n"
    "## 输出格式\n"
    "Entities:\n"
    "(实体名#实体总结)\n"
    "...\n\n"
    "Relationships:\n"
    "(来源实体名#关系名#目标实体名#关系总结)\n"
    "...\n"
    "\n"
    "## 参考案例"
    "--案例仅帮助你理解提示词的输入和输出格式，请不要在答案中使用它们。--\n"
    "输入:\n"
    "```\n"
    "[上下文]:\n"
    "Section 1:\n"
    "菲尔・贾伯的大儿子叫雅各布・贾伯。\n"
    "Section 2:\n"
    "菲尔・贾伯的小儿子叫比尔・贾伯。\n"
    "..."
    "\n"
    "[文本]:\n"
    "菲尔兹咖啡由菲尔・贾伯于1978年在加利福尼亚州伯克利创立。"
    "因其独特的混合咖啡而闻名，菲尔兹已扩展到美国多地。"
    "他的大儿子于2005年成为首席执行官，并带领公司实现了显著增长。\n"
    "```\n"
    "\n"
    "输出:\n"
    "```\n"
    "Entities:\n"
    "(菲尔・贾伯#菲尔兹咖啡创始人)\n"
    "(菲尔兹咖啡#加利福尼亚州伯克利创立的咖啡品牌)\n"
    "(雅各布・贾伯#菲尔・贾伯的大儿子)\n"
    "(美国多地#菲尔兹咖啡的扩展地区)\n"
    "\n"
    "Relationships:\n"
    "(菲尔・贾伯#创建#菲尔兹咖啡#1978年在加利福尼亚州伯克利创立)\n"
    "(菲尔兹咖啡#位于#加利福尼亚州伯克利#菲尔兹咖啡的创立地点)\n"
    "(菲尔・贾伯#拥有#雅各布・贾伯#菲尔・贾伯的大儿子)\n"
    "(雅各布・贾伯#管理#菲尔兹咖啡#在2005年担任首席执行官)\n"
    "(菲尔兹咖啡#扩展至#美国多地#菲尔兹咖啡的扩展范围)\n"
    "```\n"
    "\n"
    "----\n"
    "\n"
    "请根据接下来[上下文]提供的信息，按照上述要求，抽取[文本]中的实体和关系数据。\n"
    "\n"
    "[上下文]:\n"
    "{history}\n"
    "\n"
    "[文本]:\n"
    "{text}\n"
    "\n"
    "[结果]:\n"
    "\n"
)

GRAPH_EXTRACT_PT_EN = (
    "## Role\n"
    "You are an expert in Knowledge Graph Engineering, skilled at extracting "
    "entities (subjects, objects) and relations from text, and summarizing "
    "their meanings effectively.\n"
    "\n"
    "## Skills\n"
    "### Skill 1: Entity Extraction\n"
    "--Please follow these steps to extract entities--\n"
    "1. Accurately identify entity information in the text, "
    "usually nouns, pronouns, etc.\n"
    "2. Accurately identify descriptive information, "
    "usually as adjectives, that supplements entity features.\n"
    "3. Merge synonymous, alias, or reference entities into "
    "a single concise entity name, and merge their descriptive information.\n"
    "4. Provide a concise, appropriate, and coherent summary "
    "of the combined entity descriptions.\n"
    "\n"
    "### Skill 2: Relation Extraction\n"
    "--Please follow these steps to extract relations--\n"
    "1. Accurately identify relation information between entities in the text, "
    "usually verbs, pronouns, etc.\n"
    "2. Accurately identify descriptive information, usually as adverbs, "
    "that supplements relation features.\n"
    "3. Merge synonymous, alias, or reference relations into "
    "a single concise relation name, and merge their descriptive information.\n"
    "4. Provide a concise, appropriate, and coherent summary "
    "of the combined relation descriptions.\n"
    "\n"
    "### Skill 3: Contextual Association\n"
    "- Context comes from preceding paragraphs related to the current "
    "extraction text and can provide supplementary information.\n"
    "- Appropriately use contextual information, content references "
    "during extraction may come from this context.\n"
    "- Do not extract knowledge from contextual content, "
    "use it only as a reference.\n"
    "- Context is optional and may be empty.\n"
    "\n"
    "## Constraints\n"
    "- If the text has provided data that is similar to or the same as the "
    "output format, please format the output directly according to the "
    "output format requirements."
    "- Generate as much entity and relation information mentioned in the text "
    "as possible, but do not create nonexistent entities or relations.\n"
    "- Ensure the writing is in the third person, describing entity names, "
    "relation names, and their summaries objectively.\n"
    "- Use as much contextual information as possible to enrich the content "
    "of entities and relations, this is very important.\n"
    "- If a summary of an entity or relation is empty, do not provide "
    "summary information, and do not generate irrelevant descriptions.\n"
    "- If provided descriptions are contradictory, resolve the conflict "
    "and provide a single, coherent description.\n"
    "- Replace any # or : characters in entity's and relation's "
    "names or descriptions with an _ character.\n"
    "- Avoid using stop words and overly common terms.\n"
    "\n"
    "## Output Format\n"
    "Entities:\n"
    "(entity_name#entity_summary)\n"
    "...\n\n"
    "Relationships:\n"
    "(source_entity_name#relation_name#target_entity_name#relation_summary)\n"
    "...\n"
    "\n"
    "## Reference Example\n"
    "--The case is only to help you understand the input and output format of "
    "the prompt, please do not use it in your answer.--\n"
    "Input:\n"
    "```\n"
    "[Context]:\n"
    "Section 1:\n"
    "Phil Jabber's eldest son is named Jacob Jabber.\n"
    "Section 2:\n"
    "Phil Jabber's youngest son is named Bill Jabber.\n"
    "..."
    "\n"
    "[Text]:\n"
    "Philz Coffee was founded by Phil Jabber in 1978 in Berkeley, California. "
    "Known for its distinctive blend coffee, Philz has expanded to multiple "
    "locations in the USA. His eldest son became CEO in 2005, "
    "leading significant growth for the company.\n"
    "```\n"
    "\n"
    "Output:\n"
    "```\n"
    "Entities:\n"
    "(Phil Jabber#Founder of Philz Coffee)\n"
    "(Philz Coffee#Coffee brand founded in Berkeley, California)\n"
    "(Jacob Jabber#Phil Jabber's eldest son)\n"
    "(Multiple locations in the USA#Philz Coffee's expansion area)\n"
    "\n"
    "Relationships:\n"
    "(Phil Jabber#Founded#Philz Coffee"
    "#Founded in 1978 in Berkeley, California)\n"
    "(Philz Coffee#Located in#Berkeley, California"
    "#Philz Coffee's founding location)\n"
    "(Phil Jabber#Has#Jacob Jabber#Phil Jabber's eldest son)\n"
    "(Jacob Jabber#Manage#Philz Coffee#Serve as CEO in 2005)\n"
    "(Philz Coffee#Expanded to#Multiple locations in the USA"
    "#Philz Coffee's expansion area)\n"
    "```\n"
    "\n"
    "----\n"
    "\n"
    "Please extract the entities and relationships data from the [Text] "
    "according to the above requirements, using the provided [Context].\n"
    "\n"
    "[Context]:\n"
    "{history}\n"
    "\n"
    "[Text]:\n"
    "{text}\n"
    "\n"
    "[Results]:\n"
    "\n"
)



import unittest
from unittest.mock import Mock

from ...summary.rdbms_db_summary import RdbmsSummary


class MockRDBMSConnector(object):
    def get_users(self):
        return "user1, user2"

    def get_grants(self):
        return "grant1, grant2"

    def get_charset(self):
        return "utf8"

    def get_collation(self):
        return "utf8_general_ci"

    def get_table_names(self):
        return ["table1", "table2"]

    def get_columns(self, table_name):
        if table_name == "table1":
            return [{"name": "column1", "comment": "first column"}, {"name": "column2"}]
        return [{"name": "column1"}]

    def get_indexes(self, table_name):
        return [{"name": "index1", "column_names": ["column1"]}]

    def get_table_comment(self, table_name):
        return {"text": f"{table_name} comment"}


class TestRdbmsSummary(unittest.TestCase):
    def setUp(self):
        self.mock_local_db_manage = Mock()
        self.mock_local_db_manage.get_connector.return_value = MockRDBMSConnector()

    def test_rdbms_summary_initialization(self):
        rdbms_summary = RdbmsSummary(
            name="test_db", type="test_type", manager=self.mock_local_db_manage
        )
        self.assertEqual(rdbms_summary.name, "test_db")
        self.assertEqual(rdbms_summary.type, "test_type")
        self.assertTrue("user info :user1, user2" in rdbms_summary.metadata)
        self.assertTrue("grant info:grant1, grant2" in rdbms_summary.metadata)
        self.assertTrue("charset:utf8" in rdbms_summary.metadata)
        self.assertTrue("collation:utf8_general_ci" in rdbms_summary.metadata)

    def test_table_summaries(self):
        rdbms_summary = RdbmsSummary(
            name="test_db", type="test_type", manager=self.mock_local_db_manage
        )
        summaries = rdbms_summary.table_summaries()
        self.assertTrue(
            "table1(column1 (first column), column2), and index keys: index1(`column1`)"
            " , and table comment: table1 comment" in summaries
        )
        self.assertTrue(
            "table2(column1), and index keys: index1(`column1`) , and table comment: "
            "table2 comment" in summaries
        )


if __name__ == "__main__":
    unittest.main()



from unittest.mock import MagicMock

import pytest

from dbgpt.rag.embedding.embedding_factory import EmbeddingFactory
from dbgpt.rag.text_splitter.text_splitter import RDBTextSplitter
from dbgpt_ext.rag.assembler.db_schema import DBSchemaAssembler
from dbgpt_ext.rag.chunk_manager import ChunkParameters, SplitterType
from dbgpt_ext.storage.vector_store.chroma_store import ChromaStore


@pytest.fixture
def mock_db_connection():
    """Create a temporary database connection for testing."""
    from dbgpt_ext.datasource.rdbms.conn_sqlite import SQLiteTempConnector

    connect = SQLiteTempConnector.create_temporary_db()
    connect.create_temp_tables(
        {
            "user": {
                "columns": {
                    "id": "INTEGER PRIMARY KEY",
                    "name": "TEXT",
                    "age": "INTEGER",
                    "address": "TEXT",
                    "phone": "TEXT",
                    "email": "TEXT",
                    "gender": "TEXT",
                    "birthdate": "TEXT",
                    "occupation": "TEXT",
                    "education": "TEXT",
                    "marital_status": "TEXT",
                    "nationality": "TEXT",
                    "height": "REAL",
                    "weight": "REAL",
                    "blood_type": "TEXT",
                    "emergency_contact": "TEXT",
                    "created_at": "TEXT",
                    "updated_at": "TEXT",
                }
            }
        }
    )
    return connect


@pytest.fixture
def mock_chunk_parameters():
    return MagicMock(spec=ChunkParameters)


@pytest.fixture
def mock_embedding_factory():
    return MagicMock(spec=EmbeddingFactory)


@pytest.fixture
def mock_table_vector_store_connector():
    mock_connector = MagicMock(spec=ChromaStore)
    # mock_connector.vector_store_config.name = "table_vector_store_name"
    # mock_connector.current_embeddings = DefaultEmbeddings()
    return mock_connector


def test_load_knowledge(
    mock_db_connection,
    mock_chunk_parameters,
    mock_embedding_factory,
    mock_table_vector_store_connector,
):
    mock_chunk_parameters.chunk_strategy = "CHUNK_BY_SIZE"
    mock_chunk_parameters.text_splitter = RDBTextSplitter(
        separator="--table-field-separator--"
    )
    mock_chunk_parameters.splitter_type = SplitterType.USER_DEFINE
    assembler = DBSchemaAssembler(
        connector=mock_db_connection,
        chunk_parameters=mock_chunk_parameters,
        embeddings=mock_embedding_factory.create(),
        table_vector_store_connector=mock_table_vector_store_connector,
        max_seq_length=10,
    )
    assert len(assembler._chunks) > 1



"""Summary Assembler."""

import os
from typing import Any, List, Optional

from dbgpt.core import Chunk, LLMClient
from dbgpt.rag.knowledge.base import Knowledge
from dbgpt.rag.retriever.base import BaseRetriever
from dbgpt.rag.transformer.base import ExtractorBase

from ..assembler.base import BaseAssembler
from ..chunk_manager import ChunkParameters


class SummaryAssembler(BaseAssembler):
    """Summary Assembler.

    Example:
       .. code-block:: python

           pdf_path = "../../../DB-GPT/docs/docs/awel.md"
           OPEN_AI_KEY = "{your_api_key}"
           OPEN_AI_BASE = "{your_api_base}"
           llm_client = OpenAILLMClient(api_key=OPEN_AI_KEY, api_base=OPEN_AI_BASE)
           knowledge = KnowledgeFactory.from_file_path(pdf_path)
           chunk_parameters = ChunkParameters(chunk_strategy="CHUNK_BY_SIZE")
           assembler = SummaryAssembler.load_from_knowledge(
               knowledge=knowledge,
               chunk_parameters=chunk_parameters,
               llm_client=llm_client,
               model_name="gpt-3.5-turbo",
           )
           summary = await assembler.generate_summary()
    """

    def __init__(
        self,
        knowledge: Knowledge,
        chunk_parameters: Optional[ChunkParameters] = None,
        model_name: Optional[str] = None,
        llm_client: Optional[LLMClient] = None,
        extractor: Optional[ExtractorBase] = None,
        language: Optional[str] = "en",
        **kwargs: Any,
    ) -> None:
        """Initialize with Embedding Assembler arguments.

        Args:
            knowledge: (Knowledge) Knowledge datasource.
            chunk_manager: (Optional[ChunkManager]) ChunkManager to use for chunking.
            model_name: (Optional[str]) llm model to use.
            llm_client: (Optional[LLMClient]) LLMClient to use.
            extractor: (Optional[ExtractorBase]) ExtractorBase to use for summarization.
            language: (Optional[str]) The language of the prompt. Defaults to "en".
        """
        if knowledge is None:
            raise ValueError("knowledge datasource must be provided.")

        model_name = model_name or os.getenv("LLM_MODEL")

        if not extractor:
            from dbgpt.rag.extractor.summary import SummaryExtractor

            if not llm_client:
                raise ValueError("llm_client must be provided.")
            if not model_name:
                raise ValueError("model_name must be provided.")
            extractor = SummaryExtractor(
                llm_client=llm_client,
                model_name=model_name,
                language=language,
            )
        if not extractor:
            raise ValueError("extractor must be provided.")

        self._extractor: ExtractorBase = extractor
        super().__init__(
            knowledge=knowledge,
            chunk_parameters=chunk_parameters,
            extractor=self._extractor,
            **kwargs,
        )

    @classmethod
    def load_from_knowledge(
        cls,
        knowledge: Knowledge,
        chunk_parameters: Optional[ChunkParameters] = None,
        model_name: Optional[str] = None,
        llm_client: Optional[LLMClient] = None,
        extractor: Optional[ExtractorBase] = None,
        language: Optional[str] = "en",
        **kwargs: Any,
    ) -> "SummaryAssembler":
        """Load document embedding into vector store from path.

        Args:
            knowledge: (Knowledge) Knowledge datasource.
            chunk_parameters: (Optional[ChunkParameters]) ChunkManager to use for
                chunking.
            model_name: (Optional[str]) llm model to use.
            llm_client: (Optional[LLMClient]) LLMClient to use.
            extractor: (Optional[ExtractorBase]) ExtractorBase to use for summarization.
            language: (Optional[str]) The language of the prompt. Defaults to "en".
        Returns:
             SummaryAssembler
        """
        return cls(
            knowledge=knowledge,
            chunk_parameters=chunk_parameters,
            model_name=model_name,
            llm_client=llm_client,
            extractor=extractor,
            language=language,
            **kwargs,
        )

    async def generate_summary(self) -> str:
        """Generate summary."""
        return await self._extractor.aextract(self._chunks)

    def persist(self, **kwargs: Any) -> List[str]:
        """Persist chunks into store."""
        raise NotImplementedError

    def _extract_info(self, chunks) -> List[Chunk]:
        """Extract info from chunks."""
        return []

    def as_retriever(self, **kwargs: Any) -> BaseRetriever:
        """Return a retriever."""
        raise NotImplementedError



"""Excel Knowledge."""

import logging
from typing import Any, Dict, List, Optional, Tuple, Union

import pandas as pd

from dbgpt.core import Document
from dbgpt.rag.knowledge.base import (
    ChunkStrategy,
    DocumentType,
    Knowledge,
    KnowledgeType,
)


class ExcelKnowledge(Knowledge):
    """Excel Knowledge."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        knowledge_type: Optional[KnowledgeType] = KnowledgeType.DOCUMENT,
        source_columns: Optional[str] = None,
        encoding: Optional[str] = "utf-8",
        loader: Optional[Any] = None,
        metadata: Optional[Dict[str, Union[str, List[str]]]] = None,
        **kwargs: Any,
    ) -> None:
        """Create xlsx Knowledge with Knowledge arguments.

        Args:
            file_path(str,  optional): file path
            knowledge_type(KnowledgeType, optional): knowledge type
            source_column(str, optional): source column
            encoding(str, optional): csv encoding
            loader(Any, optional): loader
        """
        super().__init__(
            path=file_path,
            knowledge_type=knowledge_type,
            data_loader=loader,
            metadata=metadata,
            **kwargs,
        )
        self._encoding = encoding
        self._source_columns = source_columns.split(",") if source_columns else None

    def _find_header_row(
        self, data: List[List[Any]], max_rows_to_check: int = 5
    ) -> Tuple[List[str], int]:
        """
        find header row.
        """
        if not data or not data[0]:
            return [], -1

        num_cols = len(data[0])
        best_headers = []
        best_row_index = -1
        highest_score = -1

        for r_idx in range(min(len(data), max_rows_to_check)):
            row_data = data[r_idx]

            if (
                sum(1 for x in row_data if x is not None and str(x).strip() != "")
                < num_cols / 4
            ):
                continue

            potential_headers = [
                str(x).strip() if x is not None else "" for x in row_data
            ]

            cleaned_headers = [
                h
                for h in potential_headers
                if h and not h.replace(".", "", 1).isdigit()
            ]

            score = len(cleaned_headers) * 2 + len(set(cleaned_headers))

            if not any(h for h in cleaned_headers):
                continue

            avg_len = sum(len(h) for h in cleaned_headers) / (
                len(cleaned_headers) if cleaned_headers else 1
            )
            if avg_len < 2 and len(cleaned_headers) < num_cols / 2:
                continue

            if score > highest_score:
                highest_score = score
                best_headers = potential_headers
                best_row_index = r_idx

        if not best_headers:
            logging.warning(
                f"No clear header row found in the first {max_rows_to_check} rows. "
                f"Using default numeric columns."
            )
            best_headers = [str(i) for i in range(num_cols)]
            best_row_index = -1

        final_headers = []
        header_counts = {}
        for h in best_headers:
            original_h = h
            if h in header_counts:
                header_counts[h] += 1
                h = f"{original_h} ({header_counts[original_h]})"
            else:
                header_counts[h] = 0
            final_headers.append(h if h else f"Col_{len(final_headers)}")

        return final_headers, best_row_index

    def _load(self) -> List[Document]:
        """Load excel document,
        handling merged cells and intelligently identifying headers."""
        docs = []
        if not self._path:
            raise ValueError("file path is required")

        try:
            import openpyxl

            workbook = openpyxl.load_workbook(self._path)
        except Exception as e:
            raise IOError(f"Could not load Excel file with openpyxl: {e}")

        base_metadata = {
            "source": self._path,
            "data_type": "excel",
        }

        for sheet_name in workbook.sheetnames:
            sheet = workbook[sheet_name]

            max_row = sheet.max_row
            max_col = sheet.max_column

            if max_row == 0 or max_col == 0:
                logging.info(f"Sheet '{sheet_name}' is empty, skipping.")
                continue

            unmerged_data = [[None for _ in range(max_col)] for _ in range(max_row)]

            for r_idx, row_cells in enumerate(
                sheet.iter_rows(min_row=1, max_row=max_row, min_col=1, max_col=max_col)
            ):
                for c_idx, cell in enumerate(row_cells):
                    unmerged_data[r_idx][c_idx] = cell.value

            for merged_range in sheet.merged_cells.ranges:
                min_col, min_row, max_col_merged, max_row_merged = merged_range.bounds

                merged_value = sheet.cell(row=min_row, column=min_col).value

                for r in range(min_row, max_row_merged + 1):
                    for c in range(min_col, max_col_merged + 1):
                        if r - 1 < len(unmerged_data) and c - 1 < len(unmerged_data[0]):
                            unmerged_data[r - 1][c - 1] = merged_value

            headers, header_row_idx = self._find_header_row(unmerged_data)

            if header_row_idx != -1:
                data_rows = unmerged_data[header_row_idx + 1 :]
                df_start_row_index = header_row_idx + 1
                logging.info(
                    f"Sheet '{sheet_name}': Found header row at Excel "
                    f"row {header_row_idx + 1}. Data starts from Excel "
                    f"row {df_start_row_index + 1}."
                )
            else:
                data_rows = unmerged_data
                logging.info(
                    f"Sheet '{sheet_name}': No clear header row found. "
                    f"All rows considered data."
                )

            if not data_rows:
                logging.info(
                    f"Sheet '{sheet_name}': No data rows found after header "
                    f"detection, skipping."
                )
                continue

            processed_data_rows = []
            for row in data_rows:
                if len(row) < max_col:
                    processed_data_rows.append(row + [None] * (max_col - len(row)))
                elif len(row) > max_col:
                    processed_data_rows.append(row[:max_col])
                else:
                    processed_data_rows.append(row)

            df = pd.DataFrame(processed_data_rows)
            df.columns = headers

            df.dropna(axis=1, how="all", inplace=True)

            df = df.loc[:, ~df.columns.str.contains("^Unnamed:", na=False, regex=True)]

            final_headers = df.columns.tolist()
            if not final_headers:
                logging.warning(
                    f"Sheet '{sheet_name}': No valid columns remaining after "
                    f"cleanup, skipping."
                )
                continue

            for index, row in df.iterrows():
                current_metadata = base_metadata.copy()
                current_metadata["sheet_name"] = sheet_name
                current_metadata["row"] = index + df_start_row_index + 1

                strs = []

                for header_name in final_headers:
                    value = row.get(header_name)

                    if header_name is None or pd.isna(value):
                        continue

                    header_str = str(header_name).strip()
                    value_str = str(value).strip() if not pd.isna(value) else ""

                    if header_str:
                        current_metadata[header_str] = value_str

                    if self._source_columns:
                        if header_str in self._source_columns:
                            processed_value = self.parse_document_body(value_str)
                            strs.append(f"{header_str}: {processed_value}")
                    else:
                        processed_value = self.parse_document_body(value_str)
                        strs.append(f"{header_str}: {processed_value}")

                content = "\n".join(strs)

                if not content.strip():
                    logging.debug(
                        f"Skipping empty content document for "
                        f"row {current_metadata['row']} in sheet {sheet_name}"
                    )
                    continue

                if self._metadata:
                    current_metadata.update(self._metadata)

                import uuid

                if "doc_id" not in current_metadata:
                    current_metadata["doc_id"] = str(uuid.uuid4())

                doc = Document(
                    content=content,
                    metadata=current_metadata,
                )
                docs.append(doc)

            return docs

    @classmethod
    def support_chunk_strategy(cls) -> List[ChunkStrategy]:
        """Return support chunk strategy."""
        return [
            ChunkStrategy.CHUNK_BY_SIZE,
            ChunkStrategy.CHUNK_BY_SEPARATOR,
        ]

    @classmethod
    def default_chunk_strategy(cls) -> ChunkStrategy:
        """Return default chunk strategy."""
        return ChunkStrategy.CHUNK_BY_SIZE

    @classmethod
    def type(cls) -> KnowledgeType:
        """Knowledge type of CSV."""
        return KnowledgeType.DOCUMENT

    @classmethod
    def document_type(cls) -> DocumentType:
        """Return document type."""
        return DocumentType.EXCEL

    @property
    def suffix(self) -> Any:
        """Get document suffix."""
        return DocumentType.EXCEL.value



"""TripletExtractor class."""

import logging
import re
from typing import Any, List, Optional, Tuple

from dbgpt.core import LLMClient
from dbgpt.rag.transformer.llm_extractor import LLMExtractor

logger = logging.getLogger(__name__)

TRIPLET_EXTRACT_PT = (
    "Some text is provided below. Given the text, "
    "extract up to knowledge triplets as more as possible "
    "in the form of (subject, predicate, object).\n"
    "Avoid stopwords. The subject, predicate, object can not be none.\n"
    "---------------------\n"
    "Example:\n"
    "Text: Alice is Bob's mother.\n"
    "Triplets:\n(Alice, is mother of, Bob)\n"
    "Text: Alice has 2 apples.\n"
    "Triplets:\n(Alice, has 2, apple)\n"
    "Text: Alice was given 1 apple by Bob.\n"
    "Triplets:(Bob, gives 1 apple, Bob)\n"
    "Text: Alice was pushed by Bob.\n"
    "Triplets:(Bob, pushes, Alice)\n"
    "Text: Bob's mother Alice has 2 apples.\n"
    "Triplets:\n(Alice, is mother of, Bob)\n(Alice, has 2, apple)\n"
    "Text: A Big monkey climbed up the tall fruit tree and picked 3 peaches.\n"
    "Triplets:\n(monkey, climbed up, fruit tree)\n(monkey, picked 3, peach)\n"
    "Text: Alice has 2 apples, she gives 1 to Bob.\n"
    "Triplets:\n"
    "(Alice, has 2, apple)\n(Alice, gives 1 apple, Bob)\n"
    "Text: Philz is a coffee shop founded in Berkeley in 1982.\n"
    "Triplets:\n"
    "(Philz, is, coffee shop)\n(Philz, founded in, Berkeley)\n"
    "(Philz, founded in, 1982)\n"
    "---------------------\n"
    "Text: {text}\n"
    "Triplets:\n"
)


class TripletExtractor(LLMExtractor):
    """TripletExtractor class."""

    def __init__(self, llm_client: LLMClient, model_name: str):
        """Initialize the TripletExtractor."""
        super().__init__(llm_client, model_name, TRIPLET_EXTRACT_PT)

    def _parse_response(
        self, text: str, limit: Optional[int] = None
    ) -> List[Tuple[Any, ...]]:
        triplets = []

        for line in text.split("\n"):
            for match in re.findall(r"\((.*?)\)", line):
                splits = match.split(",")
                parts = [split.strip() for split in splits if split.strip()]
                if len(parts) == 3:
                    parts = [
                        p.strip(
                            "`~!@#$%^&*()-=+[]\\{}|;':\",./<>?"
                            "·！￥&*（）—【】、「」；‘’：“”，。、《》？"
                        )
                        for p in parts
                    ]
                    triplets.append(tuple(parts))
                    if limit and len(triplets) >= limit:
                        return triplets

        return triplets



"""Chroma vector store."""

import hashlib
import logging
import os
import re
from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, List, Mapping, Optional, Union

from dbgpt.configs.model_config import PILOT_PATH, resolve_root_path
from dbgpt.core import Chunk, Embeddings
from dbgpt.core.awel.flow import Parameter, ResourceCategory, register_resource
from dbgpt.storage.vector_store.base import (
    _COMMON_PARAMETERS,
    _VECTOR_STORE_COMMON_PARAMETERS,
    VectorStoreBase,
    VectorStoreConfig,
)
from dbgpt.storage.vector_store.filters import FilterOperator, MetadataFilters
from dbgpt.util import string_utils
from dbgpt.util.i18n_utils import _

logger = logging.getLogger(__name__)


@register_resource(
    _("Chroma Config"),
    "chroma_vector_config",
    category=ResourceCategory.VECTOR_STORE,
    description=_("Chroma vector store config."),
    parameters=[
        *_COMMON_PARAMETERS,
        Parameter.build_from(
            _("Persist Path"),
            "persist_path",
            str,
            description=_("the persist path of vector store."),
            optional=True,
            default=None,
        ),
    ],
)
@dataclass
class ChromaVectorConfig(VectorStoreConfig):
    """Chroma vector store config."""

    __type__ = "chroma"

    persist_path: Optional[str] = field(
        default=os.getenv("CHROMA_PERSIST_PATH", None),
        metadata={
            "help": _("The persist path of vector store."),
        },
    )
    collection_metadata: Optional[dict] = field(
        default=None,
        metadata={
            "help": _("The metadata of collection."),
        },
    )

    def create_store(self, **kwargs) -> "ChromaStore":
        """Create index store."""
        return ChromaStore(vector_store_config=self, **kwargs)


@register_resource(
    _("Chroma Vector Store"),
    "chroma_vector_store",
    category=ResourceCategory.VECTOR_STORE,
    description=_("Chroma vector store."),
    parameters=[
        Parameter.build_from(
            _("Chroma Config"),
            "vector_store_config",
            ChromaVectorConfig,
            description=_("the chroma config of vector store."),
            optional=True,
            default=None,
        ),
        *_VECTOR_STORE_COMMON_PARAMETERS,
    ],
)
class ChromaStore(VectorStoreBase):
    """Chroma vector store."""

    def __init__(
        self,
        vector_store_config: ChromaVectorConfig,
        name: Optional[str],
        embedding_fn: Optional[Embeddings] = None,
        chroma_client: Optional["PersistentClient"] = None,  # type: ignore # noqa
        collection_metadata: Optional[dict] = None,
        max_chunks_once_load: Optional[int] = None,
        max_threads: Optional[int] = None,
    ) -> None:
        """Create a ChromaStore instance.

        Args:
            vector_store_config(ChromaVectorConfig): vector store config.
            name(str): collection name.
            embedding_fn(Embeddings): embedding function.
            chroma_client(PersistentClient): chroma client.
            collection_metadata(dict): collection metadata.
            max_chunks_once_load(int): max chunks once load.
            max_threads(int): max threads.
        """
        super().__init__(
            max_chunks_once_load=max_chunks_once_load, max_threads=max_threads
        )
        self._vector_store_config = vector_store_config
        try:
            from chromadb import PersistentClient, Settings
        except ImportError:
            raise ImportError("Please install chroma package first.")
        chroma_vector_config = vector_store_config.to_dict()
        chroma_path = chroma_vector_config.get(
            "persist_path", os.path.join(PILOT_PATH, "data")
        )
        self.persist_dir = os.path.join(resolve_root_path(chroma_path) + "/chromadb")
        self.embeddings = embedding_fn
        if not self.embeddings:
            raise ValueError("Embeddings is None")
        self._collection_name = name
        if not _valid_chroma_collection_name(name):
            hash_object = hashlib.sha256(name.encode("utf-8"))
            hex_hash = hash_object.hexdigest()
            # ensure the collection name is less than 64 characters
            self._collection_name = hex_hash[:63] if len(hex_hash) > 63 else hex_hash
        chroma_settings = Settings(
            # chroma_db_impl="duckdb+parquet", => deprecated configuration of Chroma
            persist_directory=self.persist_dir,
            anonymized_telemetry=False,
        )
        self._chroma_client = chroma_client
        if not self._chroma_client:
            self._chroma_client = PersistentClient(
                path=self.persist_dir, settings=chroma_settings
            )
        collection_metadata = collection_metadata or {"hnsw:space": "cosine"}

        self._collection = self.create_collection(
            collection_name=self._collection_name,
            collection_metadata=collection_metadata,
        )

    def get_config(self) -> ChromaVectorConfig:
        """Get the vector store config."""
        return self._vector_store_config

    def create_collection(self, collection_name: str, **kwargs) -> Any:
        return self._chroma_client.get_or_create_collection(
            name=collection_name,
            embedding_function=None,
            metadata=kwargs.get("collection_metadata"),
        )

    def similar_search(
        self, text, topk, filters: Optional[MetadataFilters] = None
    ) -> List[Chunk]:
        """Search similar documents."""
        logger.info("ChromaStore similar search")
        chroma_results = self._query(
            text=text,
            topk=topk,
            filters=filters,
        )
        return [
            Chunk(
                content=chroma_result[0],
                metadata=chroma_result[1] or {},
                score=0.0,
                chunk_id=chroma_result[2],
            )
            for chroma_result in zip(
                chroma_results["documents"][0],
                chroma_results["metadatas"][0],
                chroma_results["ids"][0],
            )
        ]

    def similar_search_with_scores(
        self, text, topk, score_threshold, filters: Optional[MetadataFilters] = None
    ) -> List[Chunk]:
        """Search similar documents with scores.

        Chroma similar_search_with_score.
        Return docs and relevance scores in the range [0, 1].
        Args:
            text(str): query text
            topk(int): return docs nums. Defaults to 4.
            score_threshold(float): score_threshold: Optional, a floating point value
                between 0 to 1 to filter the resulting set of retrieved docs,0 is
                dissimilar, 1 is most similar.
            filters(MetadataFilters): metadata filters, defaults to None
        """
        logger.info("ChromaStore similar search with scores")
        chroma_results = self._query(
            text=text,
            topk=topk,
            filters=filters,
        )
        chunks = [
            (
                Chunk(
                    content=chroma_result[0],
                    metadata=chroma_result[1] or {},
                    score=(1 - chroma_result[2]),
                    chunk_id=chroma_result[3],
                )
            )
            for chroma_result in zip(
                chroma_results["documents"][0],
                chroma_results["metadatas"][0],
                chroma_results["distances"][0],
                chroma_results["ids"][0],
            )
        ]
        return self.filter_by_score_threshold(chunks, score_threshold)

    async def afull_text_search(
        self, text: str, topk: int, filters: Optional[MetadataFilters] = None
    ) -> List[Chunk]:
        """Similar search in index database.

        Args:
            text(str): The query text.
            topk(int): The number of similar documents to return.
            filters(Optional[MetadataFilters]): metadata filters.
        Return:
            List[Chunk]: The similar documents.
        """
        logger.info("ChromaStore do not support full text search")
        return []

    def is_support_full_text_search(self) -> bool:
        """Support full text search.

        Args:
            collection_name(str): collection name.
        Return:
            bool: is support full texts earch.
        """
        return False

    def vector_name_exists(self) -> bool:
        """Whether vector name exists."""
        try:
            collection = self._chroma_client.get_collection(self._collection_name)
            return collection.count() > 0
        except Exception as _e:
            logger.info(f"Collection {self._collection_name} does not exist")
            return False

    def load_document(self, chunks: List[Chunk]) -> List[str]:
        """Load document to vector store."""
        logger.info("ChromaStore load document")
        texts = [chunk.content for chunk in chunks]
        metadatas = [chunk.metadata for chunk in chunks]
        ids = [chunk.chunk_id for chunk in chunks]
        chroma_metadatas = [
            _transform_chroma_metadata(metadata) for metadata in metadatas
        ]
        self._add_texts(texts=texts, metadatas=chroma_metadatas, ids=ids)
        return ids

    def delete_vector_name(self, vector_name: str):
        """Delete vector name and clean up resources.

        Args:
            vector_name (str): Name of the vector to delete

        Returns:
            bool: True if deletion was successful, False otherwise

        Raises:
            Exception: If any error occurs during deletion
        """
        try:
            from chromadb.api.client import SharedSystemClient
        except ImportError:
            raise ImportError("Please install chroma package first.")

        logger.info(f"chroma vector_name:{vector_name} begin delete...")

        try:
            # Check if collection exists first
            collections = self._chroma_client.list_collections()
            collection_exists = self._collection.name in collections

            if not collection_exists:
                logger.warning(
                    f"Collection {self._collection.name} does not exist, skip delete"
                )
                return True

            # Delete collection if it exists
            self._chroma_client.delete_collection(self._collection.name)
            SharedSystemClient.clear_system_cache()
            return True

        except Exception as e:
            logger.error(f"Error during vector store deletion: {e}")
            raise

    def delete_by_ids(self, ids, batch_size: int = 1000):
        """Delete vector by ids in batches.

        Args:
            ids (str): Comma-separated string of IDs to delete.
            batch_size (int): Number of IDs per batch. Default is 1000.
        """
        logger.info(f"begin delete chroma ids in batches (batch size: {batch_size})")
        id_list = ids.split(",")
        total = len(id_list)
        logger.info(f"Total IDs to delete: {total}")

        for i in range(0, total, batch_size):
            batch_ids = id_list[i : i + batch_size]
            logger.info(f"Deleting batch {i // batch_size + 1}: {len(batch_ids)} IDs")
            try:
                self._collection.delete(ids=batch_ids)
            except Exception as e:
                logger.error(f"Failed to delete batch starting at index {i}: {e}")

    def truncate(self) -> List[str]:
        """Truncate data index_name."""
        logger.info(f"begin truncate chroma collection:{self._collection.name}")
        results = self._collection.get()
        ids = results.get("ids")
        if ids:
            self._collection.delete(ids=ids)
            logger.info(
                f"truncate chroma collection {self._collection.name} "
                f"{len(ids)} chunks success"
            )
            return ids
        return []

    def convert_metadata_filters(
        self,
        filters: MetadataFilters,
    ) -> dict:
        """Convert metadata filters to Chroma filters.

        Args:
            filters(MetadataFilters): metadata filters.
        Returns:
            dict: Chroma filters.
        """
        where_filters = {}
        filters_list = []
        condition = filters.condition
        chroma_condition = f"${condition.value}"
        if filters.filters:
            for filter in filters.filters:
                if filter.operator:
                    filters_list.append(
                        {
                            filter.key: {
                                _convert_chroma_filter_operator(
                                    filter.operator
                                ): filter.value
                            }
                        }
                    )
                else:
                    filters_list.append({filter.key: filter.value})  # type: ignore

        if len(filters_list) == 1:
            return filters_list[0]
        elif len(filters_list) > 1:
            where_filters[chroma_condition] = filters_list
        return where_filters

    def _add_texts(
        self,
        texts: Iterable[str],
        ids: List[str],
        metadatas: Optional[List[Mapping[str, Union[str, int, float, bool]]]] = None,
    ) -> List[str]:
        """Add texts to Chroma collection.

        Args:
            texts(Iterable[str]): texts.
            metadatas(Optional[List[dict]]): metadatas.
            ids(Optional[List[str]]): ids.
        Returns:
            List[str]: ids.
        """
        embeddings = None
        texts = list(texts)
        if self.embeddings is not None:
            embeddings = self.embeddings.embed_documents(texts)
        if metadatas:
            try:
                self._collection.upsert(
                    metadatas=metadatas,
                    embeddings=embeddings,  # type: ignore
                    documents=texts,
                    ids=ids,
                )
            except ValueError as e:
                logger.error(f"Error upsert chromadb with metadata: {e}")
        else:
            self._collection.upsert(
                embeddings=embeddings,  # type: ignore
                documents=texts,
                ids=ids,
            )
        return ids

    def _query(self, text: str, topk: int, filters: Optional[MetadataFilters] = None):
        """Query Chroma collection.

        Args:
            text(str): query text.
            topk(int): topk.
            filters(MetadataFilters): metadata filters.
        Returns:
            dict: query result.
        """
        if not text:
            return {}
        where_filters = self.convert_metadata_filters(filters) if filters else None
        if self.embeddings is None:
            raise ValueError("Chroma Embeddings is None")
        query_embedding = self.embeddings.embed_query(text)
        return self._collection.query(
            query_embeddings=query_embedding,
            n_results=topk,
            where=where_filters,
        )

    def _clean_persist_folder(self):
        """Clean persist folder."""
        for root, dirs, files in os.walk(self.persist_dir, topdown=False):
            for name in files:
                os.remove(os.path.join(root, name))
            for name in dirs:
                os.rmdir(os.path.join(root, name))
        os.rmdir(self.persist_dir)


def _convert_chroma_filter_operator(operator: str) -> str:
    """Convert operator to Chroma where operator.

    Args:
        operator(str): operator.
    Returns:
        str: Chroma where operator.
    """
    if operator == FilterOperator.EQ:
        return "$eq"
    elif operator == FilterOperator.NE:
        return "$ne"
    elif operator == FilterOperator.GT:
        return "$gt"
    elif operator == FilterOperator.LT:
        return "$lt"
    elif operator == FilterOperator.GTE:
        return "$gte"
    elif operator == FilterOperator.LTE:
        return "$lte"
    else:
        raise ValueError(f"Chroma Where operator {operator} not supported")


def _transform_chroma_metadata(
    metadata: Dict[str, Any],
) -> Mapping[str, str | int | float | bool]:
    """Transform metadata to Chroma metadata."""
    transformed = {}
    for key, value in metadata.items():
        if isinstance(value, (str, int, float, bool)):
            transformed[key] = value
    return transformed


def _valid_chroma_collection_name(name):
    """Check if the collection name is valid."""
    # ensure the collection name is less than 64 characters
    if not (3 <= len(name) <= 63):
        return False

    # ensure the collection name starts and ends with an alphanumeric character
    if not re.match(r"^[a-zA-Z0-9].*[a-zA-Z0-9]$", name):
        return False

    # ensure the collection name contains only alphanumeric characters,
    # hyphens, underscores, and dots
    if not re.match(r"^[a-zA-Z0-9_][-a-zA-Z0-9_.]*$", name):
        return False

    # ensure the collection name does not contain the '..' substring
    if ".." in name:
        return False

    if string_utils.is_valid_ipv4(name):
        return False

    if string_utils.contains_chinese(name):
        return False

    return True



"""Define Classes about Community."""

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import AsyncGenerator, Dict, Iterator, List, Optional, Union

from dbgpt.storage.graph_store.base import GraphStoreBase
from dbgpt.storage.graph_store.graph import (
    Direction,
    Edge,
    Graph,
    GraphElemType,
    MemoryGraph,
    Vertex,
)
from dbgpt.storage.knowledge_graph.base import ParagraphChunk

logger = logging.getLogger(__name__)


@dataclass
class Community:
    """Community class."""

    id: str
    data: Optional[Graph] = None
    summary: Optional[str] = None


@dataclass
class CommunityTree:
    """Represents a community tree."""


class GraphStoreAdapter(ABC):
    """Community Store Adapter."""

    def __init__(self, graph_store: GraphStoreBase):
        """Initialize Community Store Adapter."""
        self._graph_store = graph_store

    @property
    def graph_store(self) -> GraphStoreBase:
        """Get graph store."""
        return self._graph_store

    @abstractmethod
    async def discover_communities(self, **kwargs) -> List[str]:
        """Run community discovery."""

    @abstractmethod
    async def get_community(self, community_id: str) -> Community:
        """Get community."""

    @abstractmethod
    def get_graph_config(self):
        """Get config."""

    @abstractmethod
    def get_vertex_type(self) -> str:
        """Get vertex type."""

    @abstractmethod
    def get_edge_type(self) -> str:
        """Get edge type."""

    @abstractmethod
    def get_triplets(self, sub: str) -> List[tuple[str, str]]:
        """Get triplets."""

    @abstractmethod
    def get_document_vertex(self, doc_name: str) -> Vertex:
        """Get document vertex."""

    @abstractmethod
    def get_schema(self, refresh: bool = False) -> str:
        """Get schema."""

    @abstractmethod
    def get_full_graph(self, limit: Optional[int] = None) -> Graph:
        """Get full graph."""

    @abstractmethod
    def upsert_entities(self, entities: Iterator[Vertex]) -> None:
        """Upsert entity."""

    @abstractmethod
    def upsert_edge(
        self, edges: Iterator[Edge], edge_type: str, src_type: str, dst_type: str
    ):
        """Upsert edge."""

    @abstractmethod
    def upsert_chunks(
        self, chunks: Union[Iterator[Vertex], Iterator[ParagraphChunk]]
    ) -> None:
        """Upsert chunk."""

    @abstractmethod
    def upsert_documents(
        self, documents: Union[Iterator[Vertex], Iterator[ParagraphChunk]]
    ) -> None:
        """Upsert documents."""

    @abstractmethod
    def insert_triplet(self, sub: str, rel: str, obj: str) -> None:
        """Insert triplet."""

    @abstractmethod
    def upsert_graph(self, graph: Graph) -> None:
        """Insert graph."""

    @abstractmethod
    def upsert_doc_include_chunk(
        self,
        chunk: ParagraphChunk,
    ) -> None:
        """Convert chunk to document include chunk."""

    @abstractmethod
    def upsert_chunk_include_chunk(
        self,
        chunk: ParagraphChunk,
    ) -> None:
        """Convert chunk to chunk include chunk."""

    @abstractmethod
    def upsert_chunk_next_chunk(
        self,
        chunk: ParagraphChunk,
        next_chunk: ParagraphChunk,
    ):
        """Uperst the vertices and the edge in chunk_next_chunk."""

    @abstractmethod
    def upsert_chunk_include_entity(
        self, chunk: ParagraphChunk, entity: Vertex
    ) -> None:
        """Convert chunk to chunk include entity."""

    @abstractmethod
    def delete_document(self, chunk_id: str) -> None:
        """Delete document in graph store."""

    @abstractmethod
    def delete_triplet(self, sub: str, rel: str, obj: str) -> None:
        """Delete triplet."""

    @abstractmethod
    def drop(self) -> None:
        """Drop graph."""

    @abstractmethod
    def create_graph(self, graph_name: str) -> None:
        """Create graph."""

    @abstractmethod
    def create_graph_label(
        self,
        graph_elem_type: GraphElemType,
        graph_properties: List[Dict[str, Union[str, bool]]],
    ) -> None:
        """Create a graph label.

        The graph label is used to identify and distinguish different types of nodes
        (vertices) and edges in the graph.
        """

    @abstractmethod
    def truncate(self) -> None:
        """Truncate graph."""

    @abstractmethod
    def check_label(self, graph_elem_type: GraphElemType) -> bool:
        """Check if the label exists in the graph."""

    @abstractmethod
    def explore_trigraph(
        self,
        subs: Union[List[str], List[List[float]]],
        topk: Optional[int] = None,
        score_threshold: Optional[float] = None,
        direct: Direction = Direction.BOTH,
        depth: int = 3,
        fan: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> MemoryGraph:
        """Explore the graph from given subjects up to a depth.

        Args:
            subs (Union[List[str], List[List[float]]): The list of the subjects
                (keywords or embedding vectors).
            topk (Optional[int]): The number of the top similar entities.
            score_threshold (Optional[float]): The threshold of the similarity score.
            direct (Direction): The direction of the graph that will be explored.
            depth (int): The depth of the graph that will be explored.
            fan (Optional[int]): Not used.
            limit (Optional[int]): The limit number of the queried entities.

        Returns:
            MemoryGraph: The triplet graph that includes the entities and the relations.
        """

    @abstractmethod
    def explore_docgraph_with_entities(
        self,
        subs: List[str],
        topk: Optional[int] = None,
        score_threshold: Optional[float] = None,
        direct: Direction = Direction.BOTH,
        depth: int = 3,
        fan: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> MemoryGraph:
        """Explore the graph from given subjects up to a depth.

        Args:
            subs (List[str]): The list of the entities.
            topk (Optional[int]): The number of the top similar chunks.
            score_threshold (Optional[float]): The threshold of the similarity score.
            direct (Direction): The direction of the graph that will be explored.
            depth (int): The depth of the graph that will be explored.
            fan (Optional[int]): Not used.
            limit (Optional[int]): The limit number of the queried chunks.

        Returns:
            MemoryGraph: The document graph that includes the leaf chunks that connect
                to the entities, the chains from documents to the leaf chunks, and the
                chain from documents to chunks.
        """

    @abstractmethod
    def explore_docgraph_without_entities(
        self,
        subs: Union[List[str], List[List[float]]],
        topk: Optional[int] = None,
        score_threshold: Optional[float] = None,
        direct: Direction = Direction.BOTH,
        depth: int = 3,
        fan: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> MemoryGraph:
        """Explore the graph from given subjects up to a depth.

        Args:
            subs (Union[List[str], List[List[float]]): The list of the subjects
                (keywords or embedding vectors).
            topk (Optional[int]): The number of the top similar chunks.
            score_threshold (Optional[float]): The threshold of the similarity score.
            direct (Direction): The direction of the graph that will be explored.
            depth (int): The depth of the graph that will be explored.
            fan (Optional[int]): Not used.
            limit (Optional[int]): The limit number of the queried chunks.

        Returns:
            MemoryGraph: The document graph that includes the chains from documents
                to chunks that contain the subs (keywords) or similar chunks
                (embedding vectors).
        """

    @abstractmethod
    def query(self, query: str, **kwargs) -> MemoryGraph:
        """Execute a query on graph."""

    @abstractmethod
    async def stream_query(self, query: str, **kwargs) -> AsyncGenerator[Graph, None]:
        """Execute a stream query."""


class CommunityMetastore(ABC):
    """Community metastore class."""

    @abstractmethod
    def get(self, community_id: str) -> Community:
        """Get community."""

    @abstractmethod
    def list(self) -> List[Community]:
        """Get all communities."""

    @abstractmethod
    async def search(self, query: str) -> List[Community]:
        """Search communities relevant to query."""

    @abstractmethod
    async def save(self, communities: List[Community]):
        """Save communities."""

    @abstractmethod
    async def truncate(self):
        """Truncate all communities."""

    @abstractmethod
    def drop(self):
        """Drop community metastore."""


class GraphSyntaxValidator(ABC):
    """Community Syntax Validator."""

    @abstractmethod
    def validate(self, query: str) -> bool:
        """Validate query syntax."""



"""Jina Embeddings module."""

from dataclasses import dataclass, field
from typing import Any, List, Type

from dbgpt._private.pydantic import BaseModel, ConfigDict, Field
from dbgpt.core import Embeddings
from dbgpt.core.awel.flow import Parameter, ResourceCategory, register_resource
from dbgpt.model.adapter.base import register_embedding_adapter
from dbgpt.model.adapter.embed_metadata import EMBED_COMMON_HF_JINA_MODELS
from dbgpt.rag.embedding.embeddings import (
    OpenAPIEmbeddingDeployModelParameters,
    _handle_request_result,
)
from dbgpt.util.i18n_utils import _


@dataclass
class JinaEmbeddingsDeployModelParameters(OpenAPIEmbeddingDeployModelParameters):
    """Jina AI Embeddings deploy model parameters."""

    provider: str = "proxy/jina"
    api_url: str = field(
        default="https://api.jina.ai/v1/embeddings",
        metadata={
            "description": _("The URL of the embeddings API."),
            "optional": True,
        },
    )
    backend: str = field(
        default="jina-embeddings-v2-base-en",
        metadata={
            "description": _("The name of the model to use for text embeddings."),
            "optional": True,
        },
    )


@register_resource(
    _("Jina AI Embeddings"),
    "jina_embeddings",
    category=ResourceCategory.EMBEDDINGS,
    description=_("Jina AI embeddings."),
    parameters=[
        Parameter.build_from(
            _("API Key"),
            "api_key",
            str,
            description=_("Your API key for the Jina AI API."),
        ),
        Parameter.build_from(
            _("Model Name"),
            "model_name",
            str,
            optional=True,
            default="jina-embeddings-v2-base-en",
            description=_("The name of the model to use for text embeddings."),
        ),
    ],
)
class JinaEmbeddings(BaseModel, Embeddings):
    """Jina AI embeddings.

    This class is used to get embeddings for a list of texts using the Jina AI API.
    It requires an API key and a model name. The default model name is
    "jina-embeddings-v2-base-en".
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, protected_namespaces=())

    api_url: Any  #: :meta private:
    session: Any  #: :meta private:
    api_key: str
    timeout: int = Field(
        default=60, description="The timeout for the request in seconds."
    )
    """API key for the Jina AI API.."""
    model_name: str = "jina-embeddings-v2-base-en"
    """The name of the model to use for text embeddings. Defaults to
    "jina-embeddings-v2-base-en"."""

    def __init__(self, **kwargs):
        """Create a new JinaEmbeddings instance."""
        try:
            import requests
        except ImportError:
            raise ValueError(
                "The requests python package is not installed. Please install it with "
                "`pip install requests`"
            )
        if "api_url" not in kwargs:
            kwargs["api_url"] = "https://api.jina.ai/v1/embeddings"
        if "session" not in kwargs:  # noqa: SIM401
            session = requests.Session()
        else:
            session = kwargs["session"]
        api_key = kwargs.get("api_key")
        if api_key:
            session.headers.update(
                {
                    "Authorization": f"Bearer {api_key}",
                    "Accept-Encoding": "identity",
                }
            )
        kwargs["session"] = session

        super().__init__(**kwargs)

    @classmethod
    def param_class(cls) -> Type[JinaEmbeddingsDeployModelParameters]:
        """Get the parameter class."""
        return JinaEmbeddingsDeployModelParameters

    @classmethod
    def from_parameters(
        cls, parameters: JinaEmbeddingsDeployModelParameters
    ) -> "Embeddings":
        """Create an embedding model from parameters."""
        return cls(
            api_url=parameters.api_url,
            api_key=parameters.api_key,
            model_name=parameters.real_provider_model_name,
            timeout=parameters.timeout,
        )

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Get the embeddings for a list of texts.

        Args:
            texts (Documents): A list of texts to get embeddings for.

        Returns:
            Embedded texts as List[List[float]], where each inner List[float]
                corresponds to a single input text.
        """
        # Call Jina AI Embedding API
        resp = self.session.post(  # type: ignore
            self.api_url,
            json={"input": texts, "model": self.model_name},
            timeout=self.timeout,
        )
        return _handle_request_result(resp)

    def embed_query(self, text: str) -> List[float]:
        """Compute query embeddings using a Jina AI embedding model.

        Args:
            text: The text to embed.

        Returns:
            Embeddings for the text.
        """
        return self.embed_documents([text])[0]


register_embedding_adapter(JinaEmbeddings, supported_models=EMBED_COMMON_HF_JINA_MODELS)



"""DBSchema retriever."""

import logging
from typing import List, Optional

from dbgpt._private.config import Config
from dbgpt.core import Chunk
from dbgpt.datasource.base import BaseConnector
from dbgpt.rag.retriever.base import BaseRetriever
from dbgpt.rag.retriever.rerank import DefaultRanker, Ranker
from dbgpt.storage.vector_store.base import VectorStoreBase
from dbgpt.storage.vector_store.filters import MetadataFilter, MetadataFilters
from dbgpt.util.chat_util import run_tasks
from dbgpt.util.executor_utils import blocking_func_to_async_no_executor

from ..summary.rdbms_db_summary import (
    _DEFAULT_COLUMN_SEPARATOR,
    _parse_db_summary,
    _parse_table_detail,
)

logger = logging.getLogger(__name__)

CFG = Config()


class DBSchemaRetriever(BaseRetriever):
    """DBSchema retriever."""

    def __init__(
        self,
        table_vector_store_connector: VectorStoreBase,
        field_vector_store_connector: VectorStoreBase = None,
        separator: str = "--table-field-separator--",
        column_separator: str = _DEFAULT_COLUMN_SEPARATOR,
        top_k: int = 4,
        connector: Optional[BaseConnector] = None,
        query_rewrite: bool = False,
        rerank: Optional[Ranker] = None,
        **kwargs,
    ):
        """Create DBSchemaRetriever.

        Args:
            table_vector_store_connector: VectorStoreBase
                to load and retrieve table info.
            field_vector_store_connector: VectorStoreBase
                to load and retrieve field info.
            separator: field/table separator
            top_k (int): top k
            connector (Optional[BaseConnector]): RDBMSConnector.
            query_rewrite (bool): query rewrite
            rerank (Ranker): rerank

        Examples:
            .. code-block:: python

                from dbgpt_ext.datasource.rdbms.conn_sqlite import SQLiteTempConnector
                from dbgpt_ext.rag.assembler.db_schema import DBSchemaAssembler
                from dbgpt_serve.rag.connector import VectorStoreConnector
                from dbgpt_ext.storage.vector_store.chroma_store import (
                    ChromaVectorConfig,
                )
                from dbgpt.rag.retriever.embedding import EmbeddingRetriever


                def _create_temporary_connection():
                    connect = SQLiteTempConnector.create_temporary_db()
                    connect.create_temp_tables(
                        {
                            "user": {
                                "columns": {
                                    "id": "INTEGER PRIMARY KEY",
                                    "name": "TEXT",
                                    "age": "INTEGER",
                                },
                                "data": [
                                    (1, "Tom", 10),
                                    (2, "Jerry", 16),
                                    (3, "Jack", 18),
                                    (4, "Alice", 20),
                                    (5, "Bob", 22),
                                ],
                            }
                        }
                    )
                    return connect


                connector = _create_temporary_connection()
                vector_store_config = ChromaVectorConfig(name="vector_store_name")
                embedding_model_path = "{your_embedding_model_path}"
                embedding_fn = embedding_factory.create(model_name=embedding_model_path)
                vector_connector = VectorStoreConnector.from_default(
                    "Chroma",
                    vector_store_config=vector_store_config,
                    embedding_fn=embedding_fn,
                )
                # get db struct retriever
                retriever = DBSchemaRetriever(
                    top_k=3,
                    vector_store_connector=vector_connector,
                    connector=connector,
                )
                chunks = retriever.retrieve("show columns from table")
                result = [chunk.content for chunk in chunks]
                print(f"db struct rag example results:{result}")
        """
        self._separator = separator
        self._column_separator = column_separator
        self._top_k = top_k
        self._connector = connector
        self._query_rewrite = query_rewrite
        self._table_vector_store_connector = table_vector_store_connector
        self._field_vector_store_connector = field_vector_store_connector
        self._need_embeddings = False
        if self._table_vector_store_connector:
            self._need_embeddings = True
        self._rerank = rerank or DefaultRanker(self._top_k)

    def _retrieve(
        self, query: str, filters: Optional[MetadataFilters] = None
    ) -> List[Chunk]:
        """Retrieve knowledge chunks.

        Args:
            query (str): query text
            filters: metadata filters.

        Returns:
            List[Chunk]: list of chunks
        """
        if self._need_embeddings:
            return self._similarity_search(query, filters)
        else:
            table_summaries = _parse_db_summary(self._connector)
            return [Chunk(content=table_summary) for table_summary in table_summaries]

    def _retrieve_with_score(
        self,
        query: str,
        score_threshold: float,
        filters: Optional[MetadataFilters] = None,
    ) -> List[Chunk]:
        """Retrieve knowledge chunks with score.

        Args:
            query (str): query text
            score_threshold (float): score threshold
            filters: metadata filters.

        Returns:
            List[Chunk]: list of chunks
        """
        return self._retrieve(query, filters)

    async def _aretrieve(
        self, query: str, filters: Optional[MetadataFilters] = None
    ) -> List[Chunk]:
        """Retrieve knowledge chunks.

        Args:
            query (str): query text
            filters: metadata filters.

        Returns:
            List[Chunk]: list of chunks
        """
        return await blocking_func_to_async_no_executor(
            func=self._retrieve,
            query=query,
            filters=filters,
        )

    async def _aretrieve_with_score(
        self,
        query: str,
        score_threshold: float,
        filters: Optional[MetadataFilters] = None,
    ) -> List[Chunk]:
        """Retrieve knowledge chunks with score.

        Args:
            query (str): query text
            score_threshold (float): score threshold
            filters: metadata filters.
        """
        return await self._aretrieve(query, filters)

    def _retrieve_field(self, table_chunk: Chunk, query) -> Chunk:
        metadata = table_chunk.metadata
        metadata["part"] = "field"
        filters = [MetadataFilter(key=k, value=v) for k, v in metadata.items()]
        field_chunks = self._field_vector_store_connector.similar_search_with_scores(
            query, self._top_k, 0, MetadataFilters(filters=filters)
        )
        field_contents = [chunk.content.strip() for chunk in field_chunks]
        table_chunk.content += (
            "\n" + self._separator + "\n" + self._column_separator.join(field_contents)
        )
        return self._deserialize_table_chunk(table_chunk)

    def _similarity_search(
        self, query, filters: Optional[MetadataFilters] = None
    ) -> List[Chunk]:
        """Similar search."""
        table_chunks = self._table_vector_store_connector.similar_search_with_scores(
            query, self._top_k, 0, filters
        )

        # Find all table chunks which are not separated
        not_sep_chunks = [
            chunk for chunk in table_chunks if not chunk.metadata.get("separated")
        ]
        separated_chunks = [
            chunk for chunk in table_chunks if chunk.metadata.get("separated")
        ]
        if not separated_chunks:
            return [self._deserialize_table_chunk(chunk) for chunk in not_sep_chunks]

        # Create tasks list
        # The fields of table is too large, and it has to be separated into chunks,
        # so we need to retrieve fields of each table separately
        tasks = [
            lambda c=chunk: self._retrieve_field(c, query) for chunk in separated_chunks
        ]
        # Run tasks concurrently
        separated_result = run_tasks(tasks, concurrency_limit=3)

        # Combine and return results
        return not_sep_chunks + separated_result

    def _deserialize_table_chunk(self, chunk: Chunk) -> Chunk:
        """Deserialize table chunk."""
        db_summary_version = chunk.metadata.get("db_summary_version")
        if not db_summary_version:
            return chunk
        parts = chunk.content.split(self._separator)
        table_part, field_part = parts[0].strip(), parts[1].strip()
        table_detail = _parse_table_detail(table_part)
        table_name = table_detail.get("table_name")
        table_comment = table_detail.get("table_comment")
        index_keys = table_detail.get("index_keys")

        table_name = table_name.strip() if table_name else table_name
        table_comment = table_comment.strip() if table_comment else table_comment
        index_keys = index_keys.strip() if index_keys else index_keys
        if not table_name:
            return chunk

        create_statement = f"CREATE TABLE `{table_name}`\r\n(\r\n    "
        create_statement += field_part
        create_statement += "\r\n)"
        if table_comment:
            create_statement += f' COMMENT "{table_comment}"\r\n'
        if index_keys:
            create_statement += f"Index keys: {index_keys}"

        chunk.content = create_statement
        return chunk



"""Graph retriever base class."""

import logging
from abc import ABC, abstractmethod
from typing import Any, Tuple

from dbgpt.storage.graph_store.graph import Graph

logger = logging.getLogger(__name__)


class GraphRetrieverBase(ABC):
    """Graph retriever base class."""

    @abstractmethod
    async def retrieve(self, input: Any) -> Tuple[Graph, Any]:
        """Retrieve from graph database."""



"""Markdown Knowledge."""

from typing import Any, Dict, List, Optional, Union

from dbgpt.core import Document
from dbgpt.rag.knowledge.base import (
    ChunkStrategy,
    DocumentType,
    Knowledge,
    KnowledgeType,
)
from dbgpt_ext.rag import ChunkParameters


class MarkdownKnowledge(Knowledge):
    """Markdown Knowledge."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        knowledge_type: KnowledgeType = KnowledgeType.DOCUMENT,
        encoding: Optional[str] = "utf-8",
        loader: Optional[Any] = None,
        metadata: Optional[Dict[str, Union[str, List[str]]]] = None,
        **kwargs: Any,
    ) -> None:
        """Create Markdown Knowledge with Knowledge arguments.

        Args:
            file_path(str,  optional): file path
            knowledge_type(KnowledgeType, optional): knowledge type
            encoding(str, optional): csv encoding
            loader(Any, optional): loader
        """
        super().__init__(
            path=file_path,
            knowledge_type=knowledge_type,
            data_loader=loader,
            metadata=metadata,
            **kwargs,
        )
        self._encoding = encoding

    def _load(self) -> List[Document]:
        """Load markdown document from loader."""
        if self._loader:
            documents = self._loader.load()
        else:
            if not self._path:
                raise ValueError("file path is required")
            with open(self._path, encoding=self._encoding, errors="ignore") as f:
                markdown_text = f.read()
                metadata = {
                    "source": self._path,
                    "title": self._path.rsplit("/", 1)[-1],
                }
                if self._metadata:
                    metadata.update(self._metadata)  # type: ignore
                documents = [Document(content=markdown_text, metadata=metadata)]
                return documents
        return [Document.langchain2doc(lc_document) for lc_document in documents]

    def extract(
        self,
        documents: List[Document],
        chunk_parameter: Optional[ChunkParameters] = None,
    ) -> List[Document]:
        """Extract knowledge from text."""
        from dbgpt_ext.rag.chunk_manager import ChunkManager

        chunk_manager = ChunkManager(knowledge=self, chunk_parameter=chunk_parameter)
        chunks = chunk_manager.split(documents)
        for document in documents:
            document.chunks = chunks
        return documents

    @classmethod
    def support_chunk_strategy(cls) -> List[ChunkStrategy]:
        """Return support chunk strategy."""
        return [
            ChunkStrategy.CHUNK_BY_SIZE,
            ChunkStrategy.CHUNK_BY_MARKDOWN_HEADER,
            ChunkStrategy.CHUNK_BY_SEPARATOR,
        ]

    @classmethod
    def default_chunk_strategy(cls) -> ChunkStrategy:
        """Return default chunk strategy."""
        return ChunkStrategy.CHUNK_BY_MARKDOWN_HEADER

    @classmethod
    def type(cls) -> KnowledgeType:
        """Return knowledge type."""
        return KnowledgeType.DOCUMENT

    @classmethod
    def document_type(cls) -> DocumentType:
        """Return document type."""
        return DocumentType.MARKDOWN



"""Tongyi embeddings for RAG."""

from dataclasses import dataclass, field
from typing import List, Optional, Type

from dbgpt._private.pydantic import BaseModel, ConfigDict, Field
from dbgpt.core import EmbeddingModelMetadata, Embeddings
from dbgpt.core.interface.parameter import EmbeddingDeployModelParameters
from dbgpt.model.adapter.base import register_embedding_adapter
from dbgpt.util.i18n_utils import _


@dataclass
class TongyiEmbeddingDeployModelParameters(EmbeddingDeployModelParameters):
    """Qianfan Embeddings deploy model parameters."""

    provider: str = "proxy/tongyi"

    api_key: Optional[str] = field(
        default=None, metadata={"help": _("The API key for the embeddings API.")}
    )
    backend: Optional[str] = field(
        default="text-embedding-v1",
        metadata={
            "help": _(
                "The real model name to pass to the provider, default is None. If "
                "backend is None, use name as the real model name."
            ),
        },
    )

    @property
    def real_provider_model_name(self) -> str:
        """Get the real provider model name."""
        return self.backend or self.name


class TongYiEmbeddings(BaseModel, Embeddings):
    """The tongyi embeddings.

    import dashscope
    from http import HTTPStatus
    from dashscope import TextEmbedding

    dashscope.api_key = ''
    def embed_with_list_of_str():
        resp = TextEmbedding.call(
            model=TextEmbedding.Models.text_embedding_v1,
            # 最多支持10条，每条最长支持2048tokens
            input=[
                '风急天高猿啸哀', '渚清沙白鸟飞回', '无边落木萧萧下', '不尽长江滚滚来'
            ]
        )
        if resp.status_code == HTTPStatus.OK:
            print(resp)
        else:
            print(resp)

    if __name__ == '__main__':
        embed_with_list_of_str()
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, protected_namespaces=())
    api_key: Optional[str] = Field(
        default=None, description="The API key for the embeddings API."
    )
    model_name: str = Field(
        default="text-embedding-v1", description="The name of the model to use."
    )

    def __init__(self, **kwargs):
        """Initialize the OpenAPIEmbeddings."""
        try:
            import dashscope  # type: ignore
        except ImportError as exc:
            raise ValueError(
                "Could not import python package: dashscope "
                "Please install dashscope by command `pip install dashscope"
            ) from exc
        dashscope.TextEmbedding.api_key = kwargs.get("api_key")
        super().__init__(**kwargs)
        self._api_key = kwargs.get("api_key")

    @classmethod
    def param_class(cls) -> Type[TongyiEmbeddingDeployModelParameters]:
        """Get the parameter class."""
        return TongyiEmbeddingDeployModelParameters

    @classmethod
    def from_parameters(
        cls, parameters: TongyiEmbeddingDeployModelParameters
    ) -> "Embeddings":
        """Create an embedding model from parameters."""
        return cls(
            api_key=parameters.api_key,
            model_name=parameters.real_provider_model_name,
        )

    def embed_documents(
        self, texts: List[str], max_batch_chunks_size=25
    ) -> List[List[float]]:
        """Get the embeddings for a list of texts.

        refer:https://help.aliyun.com/zh/model-studio/getting-started/models?
        spm=a2c4g.11186623.0.0.62524a77NlILDI#c05fe72732770

        Args:
            texts (Documents): A list of texts to get embeddings for.
            max_batch_chunks_size: The max batch size for embedding.

        Returns:
            Embedded texts as List[List[float]], where each inner List[float]
                corresponds to a single input text.
        """
        from dashscope import TextEmbedding

        embeddings = []
        # batch size too longer may cause embedding error,eg: qwen online embedding
        # models must not be larger than 25
        # text-embedding-v3  embedding batch size should not be larger than 6
        if str(self.model_name) == "text-embedding-v3":
            max_batch_chunks_size = 6

        for i in range(0, len(texts), max_batch_chunks_size):
            batch_texts = texts[i : i + max_batch_chunks_size]
            resp = TextEmbedding.call(
                model=self.model_name, input=batch_texts, api_key=self._api_key
            )
            if "output" not in resp:
                raise RuntimeError(resp["message"])

            # 提取并排序嵌入
            batch_embeddings = resp["output"]["embeddings"]
            sorted_embeddings = sorted(batch_embeddings, key=lambda e: e["text_index"])
            embeddings.extend([result["embedding"] for result in sorted_embeddings])

        return embeddings

    def embed_query(self, text: str) -> List[float]:
        """Compute query embeddings using a OpenAPI embedding model.

        Args:
            text: The text to embed.

        Returns:
            Embeddings for the text.
        """
        return self.embed_documents([text])[0]


register_embedding_adapter(
    TongYiEmbeddings,
    supported_models=[
        EmbeddingModelMetadata(
            model="text-embedding-v3",
            dimension=1024,
            context_length=8192,
            description=_(
                "The embedding model are trained by TongYi, it support more than 50 "
                "working languages."
            ),
        )
    ],
)



"""PDF Knowledge."""

import json
import os
import re
from collections import defaultdict
from typing import Any, Dict, List, Optional, Union

from dbgpt.component import logger
from dbgpt.core import Document
from dbgpt.rag.knowledge.base import (
    ChunkStrategy,
    DocumentType,
    Knowledge,
    KnowledgeType,
)


class PDFKnowledge(Knowledge):
    """PDF Knowledge."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        knowledge_type: KnowledgeType = KnowledgeType.DOCUMENT,
        loader: Optional[Any] = None,
        language: Optional[str] = "zh",
        metadata: Optional[Dict[str, Union[str, List[str]]]] = None,
        **kwargs: Any,
    ) -> None:
        """Create PDF Knowledge with Knowledge arguments.

        Args:
            file_path(str,  optional): file path
            knowledge_type(KnowledgeType, optional): knowledge type
            loader(Any, optional): loader
            language(str, optional): language
        """
        super().__init__(
            path=file_path,
            knowledge_type=knowledge_type,
            data_loader=loader,
            metadata=metadata,
            **kwargs,
        )
        self._language = language
        self._pdf_processor = PDFProcessor(filepath=self._path)
        self.all_title: List[dict] = []
        self.all_text: List[dict] = []

    def process_text_data(self):
        """Text data processing to level 1 and level 2 titles."""
        for i, data in enumerate(self.all_text):
            # data = self.all_text[i]
            inside_content = data.get("inside")
            content_type = data.get("type")
            if content_type == "text":
                # use regex to match the first level title
                first_level_match = re.match(
                    r"§(\d+)+([\u4e00-\u9fa5]+)", inside_content.strip()
                )
                second_level_match = re.match(
                    r"(\d+\.\d+)([\u4e00-\u9fa5]+)", inside_content.strip()
                )
                first_num_match = re.match(r"^§(\d+)$", inside_content.strip())
                # get all level 1 titles
                title_name = [
                    dictionary["first_title"]
                    for dictionary in self.all_title
                    if "first_title" in dictionary
                ]
                if first_level_match:
                    first_title_text = first_level_match.group(2)
                    first_title_num = first_level_match.group(1)
                    first_title = first_title_num + first_title_text
                    # the title does not contain "..." and is not in the title list
                    # , add it to the title list
                    if first_title not in title_name and (
                        int(first_title_num) == 1
                        or int(first_title_num) - int(self.all_title[-1]["id"]) == 1
                    ):
                        current_entry = {
                            "id": first_title_num,
                            "first_title": first_title,
                            "second_title": [],
                            "table": [],
                        }
                        self.all_title.append(current_entry)

                elif second_level_match:
                    second_title_name = second_level_match.group(0)
                    second_title = second_level_match.group(1)
                    first_title = second_title.split(".")[0]
                    if (int(first_title) - 1 >= len(self.all_title)) or int(
                        first_title
                    ) - 1 < 0:
                        continue
                    else:
                        titles = [
                            sub_item["title"]
                            for sub_item in self.all_title[int(first_title) - 1][
                                "second_title"
                            ]
                        ]
                        if second_title_name not in titles:
                            self.all_title[int(first_title) - 1]["second_title"].append(
                                {"title": second_title_name, "table": []}
                            )
                elif first_num_match:
                    first_num = first_num_match.group(1)
                    first_text = self.all_text[i + 1].get("inside")
                    first_title = first_num_match.group(1) + first_text
                    # if the title does not contain "..." and is not in the title list
                    if (
                        "..." not in first_text
                        and first_title not in title_name
                        and (
                            int(first_num) == 1
                            or int(first_num) - int(self.all_title[-1]["id"]) == 1
                        )
                    ):
                        current_entry = {
                            "id": first_num,
                            "first_title": first_title,
                            "second_title": [],
                            "table": [],
                        }
                        self.all_title.append(current_entry)

    def _load(self) -> List[Document]:
        """Load pdf document from loader."""
        if self._loader:
            documents = self._loader.load()
        else:
            self._pdf_processor.pdf_to_json()
            file_title = self.file_path.rsplit("/", 1)[-1].replace(".pdf", "")
            self.all_text = list(self._pdf_processor.all_text.values())
            self.process_text_data()
            temp_table = []
            temp_title = None
            page_documents = []
            merged_data = {}  # type: ignore # noqa
            for i, data in enumerate(self.all_text):
                content_type = data.get("type")
                inside_content = data.get("inside")
                page = data.get("page")

                if content_type == "excel":
                    temp_table.append(inside_content)
                    if temp_title is None:
                        for j in range(i - 1, -1, -1):
                            if self.all_text[j]["type"] == "excel":
                                break
                            if self.all_text[j]["type"] == "text":
                                content = self.all_text[j]["inside"]
                                if re.match(
                                    r"^\d+\.\d+", content
                                ) or content.startswith("§"):
                                    temp_title = content.strip()
                                    break
                                else:
                                    temp_title = content.strip()
                                    break
                elif content_type == "text":
                    if page in merged_data:
                        # page merge
                        merged_data[page]["inside_content"] += " " + inside_content
                    else:
                        merged_data[page] = {
                            "inside_content": inside_content,
                            "type": "text",
                        }

                    # merge excel table
                    if temp_table:
                        table_meta = {
                            "title": temp_title or temp_table[0],
                            "type": "excel",
                        }
                        self.all_title.append(table_meta)

                        # markdown format
                        markdown_tables = []
                        if temp_table:
                            header = eval(temp_table[0])
                            markdown_tables.append(header)
                            for entry in temp_table[1:]:
                                row = eval(entry)
                                markdown_tables.append(row)
                            markdown_output = "| " + " | ".join(header) + " |\n"
                            markdown_output += (
                                "| " + " | ".join(["---"] * len(header)) + " |\n"
                            )
                            for row in markdown_tables[1:]:
                                markdown_output += "| " + " | ".join(row) + " |\n"

                            #  merged content
                            merged_data[page]["excel_content"] = temp_table
                            merged_data[page]["markdown_output"] = markdown_output

                        temp_title = None
                        temp_table = []

            # deal last excel
            if temp_table:
                table_meta = {
                    "title": temp_title or temp_table[0],
                    "table": temp_table,
                    "type": "excel",
                }
                self.all_title.append(table_meta)
                # markdown format
                markdown_tables = []
                if temp_table:
                    header = eval(temp_table[0])
                    markdown_tables.append(header)
                    for entry in temp_table[1:]:
                        row = eval(entry)
                        markdown_tables.append(row)
                    markdown_output = "| " + " | ".join(header) + " |\n"
                    markdown_output += "| " + " | ".join(["---"] * len(header)) + " |\n"
                    for row in markdown_tables[1:]:
                        markdown_output += "| " + " | ".join(row) + " |\n"
                    #  merged content
                    merged_data[page]["excel_content"] = temp_table
                    merged_data[page]["markdown_output"] = markdown_output

            for page, content in merged_data.items():
                inside_content = content["inside_content"]
                if "markdown_output" in content:
                    markdown_content = content["markdown_output"]
                    content_metadata = {
                        "page": page,
                        "type": "excel",
                        "title": file_title,
                        "source": self.file_path,
                    }
                    page_documents.append(
                        Document(
                            content=inside_content + "\n" + markdown_content,
                            metadata=content_metadata,
                        )
                    )
                else:
                    content_metadata = {
                        "page": page,
                        "type": "text",
                        "title": file_title,
                        "source": self.file_path,
                    }
                    page_documents.append(
                        Document(content=inside_content, metadata=content_metadata)
                    )

            return page_documents
        return [Document.langchain2doc(lc_document) for lc_document in documents]

    @classmethod
    def support_chunk_strategy(cls) -> List[ChunkStrategy]:
        """Return support chunk strategy."""
        return [
            ChunkStrategy.CHUNK_BY_SIZE,
            ChunkStrategy.CHUNK_BY_PAGE,
            ChunkStrategy.CHUNK_BY_SEPARATOR,
        ]

    @classmethod
    def default_chunk_strategy(cls) -> ChunkStrategy:
        """Return default chunk strategy."""
        return ChunkStrategy.CHUNK_BY_SIZE

    @classmethod
    def type(cls) -> KnowledgeType:
        """Return knowledge type."""
        return KnowledgeType.DOCUMENT

    @classmethod
    def document_type(cls) -> DocumentType:
        """Document type of PDF."""
        return DocumentType.PDF


class PDFProcessor:
    """PDFProcessor class."""

    def __init__(self, filepath):
        """Initialize PDFProcessor class."""
        self.filepath = filepath
        try:
            import pdfplumber  # type: ignore
        except ImportError:
            raise ImportError("Please install pdfplumber first.")
        self.pdf = pdfplumber.open(filepath)
        self.all_text = defaultdict(dict)
        self.allrow = 0
        self.last_num = 0

    def check_lines(self, page, top, buttom):
        """Check lines."""
        lines = page.extract_words()[::]
        text = ""
        last_top = 0
        last_check = 0
        for line in range(len(lines)):
            each_line = lines[line]
            check_re = (
                "(?:。|；|单位：人民币元|金额单位：人民币元|单位：万元|币种：人民币|\d|"
                "报告(?:全文)?(?:（修订版）|（修订稿）|（更正后）)?)$"
            )
            if top == "" and buttom == "":
                if abs(last_top - each_line["top"]) <= 2 or (
                    last_check > 0
                    and (page.height * 0.9 - each_line["top"]) > 0
                    and not re.search(check_re, text)
                ):
                    text = text + each_line["text"]
                else:
                    text = text + "\n" + each_line["text"]
            elif top == "":
                if each_line["top"] > buttom:
                    if abs(last_top - each_line["top"]) <= 2 or (
                        last_check > 0
                        and (page.height * 0.85 - each_line["top"]) > 0
                        and not re.search(check_re, text)
                    ):
                        text = text + each_line["text"]
                    else:
                        text = text + "\n" + each_line["text"]
            else:
                if each_line["top"] < top and each_line["top"] > buttom:
                    if abs(last_top - each_line["top"]) <= 2 or (
                        last_check > 0
                        and (page.height * 0.85 - each_line["top"]) > 0
                        and not re.search(check_re, text)
                    ):
                        text = text + each_line["text"]
                    else:
                        text = text + "\n" + each_line["text"]
            last_top = each_line["top"]
            last_check = each_line["x1"] - page.width * 0.85

        return text

    def drop_empty_cols(self, data):
        """Delete empty column."""
        transposed_data = list(map(list, zip(*data)))
        filtered_data = [
            col for col in transposed_data if not all(cell == "" for cell in col)
        ]
        result = list(map(list, zip(*filtered_data)))
        return result

    def extract_text_and_tables(self, page):
        """Extract text and tables."""
        buttom = 0
        tables = page.find_tables()
        if len(tables) >= 1:
            count = len(tables)
            for table in tables:
                # process text before table
                if table.bbox[3] < buttom:
                    pass
                else:
                    count -= 1
                    # process text before table
                    top = table.bbox[1]
                    text = self.check_lines(page, top, buttom)
                    text_list = text.split("\n")
                    for _t in range(len(text_list)):
                        self.all_text[self.allrow] = {
                            "page": page.page_number,
                            "allrow": self.allrow,
                            "type": "text",
                            "inside": text_list[_t],
                        }
                        self.allrow += 1

                    # process table
                    buttom = table.bbox[3]
                    new_table = table.extract()
                    r_count = 0
                    for r in range(len(new_table)):
                        row = new_table[r]
                        if row[0] is None:
                            r_count += 1
                            for c in range(len(row)):
                                if row[c] is not None and row[c] not in ["", " "]:
                                    if new_table[r - r_count][c] is None:
                                        new_table[r - r_count][c] = row[c]
                                    else:
                                        new_table[r - r_count][c] += row[c]
                                    new_table[r][c] = None
                        else:
                            r_count = 0

                    end_table = []
                    for row in new_table:
                        if row[0] is not None:
                            cell_list = []
                            cell_check = False
                            for cell in row:
                                if cell is not None:
                                    cell = cell.replace("\n", "")
                                else:
                                    cell = ""
                                if cell != "":
                                    cell_check = True
                                cell_list.append(cell)
                            if cell_check:
                                end_table.append(cell_list)

                    end_table = self.drop_empty_cols(end_table)

                    # process when column name is empty
                    if len(end_table) > 0:
                        for i in range(len(end_table[0])):
                            if end_table[0][i] == "":
                                if 0 < i < len(end_table[0]) - 1:
                                    # left column name
                                    left_column = end_table[0][i - 1]
                                    # right column name
                                    right_column = end_table[0][i + 1]
                                    # current name = left name + right name
                                    end_table[0][i] = left_column + right_column
                                else:
                                    # if current column is empty and is the first
                                    # column, assign the right column name.
                                    # if current column is empty and is the
                                    # last column, assign the left column name.
                                    end_table[0][i] = (
                                        end_table[0][i - 1]
                                        if i == len(end_table[0]) - 1
                                        else end_table[0][i + 1]
                                    )

                    # if the first row is empty, assign the value of the previous row
                    for i in range(1, len(end_table)):
                        for j in range(len(end_table[i])):
                            if end_table[i][j] == "":
                                end_table[i][j] = end_table[i][j - 1]

                    for row in end_table:
                        self.all_text[self.allrow] = {
                            "page": page.page_number,
                            "allrow": self.allrow,
                            "type": "excel",
                            "inside": str(row),
                        }
                        self.allrow += 1

                    if count == 0:
                        text = self.check_lines(page, "", buttom)
                        text_list = text.split("\n")
                        for _t in range(len(text_list)):
                            self.all_text[self.allrow] = {
                                "page": page.page_number,
                                "allrow": self.allrow,
                                "type": "text",
                                "inside": text_list[_t],
                            }
                            self.allrow += 1

        else:
            text = self.check_lines(page, "", "")
            text_list = text.split("\n")
            for _t in range(len(text_list)):
                self.all_text[self.allrow] = {
                    "page": page.page_number,
                    "allrow": self.allrow,
                    "type": "text",
                    "inside": text_list[_t],
                }
                self.allrow += 1

        first_re = "[^计](?:报告(?:全文)?(?:（修订版）|（修订稿）|（更正后）)?)$"
        end_re = "^(?:\d|\\|\/|第|共|页|-|_| ){1,}"
        if self.last_num == 0:
            try:
                first_text = str(self.all_text[1]["inside"])
                end_text = str(self.all_text[len(self.all_text) - 1]["inside"])
                if re.search(first_re, first_text) and "[" not in end_text:
                    self.all_text[1]["type"] = "页眉"
                    if re.search(end_re, end_text) and "[" not in end_text:
                        self.all_text[len(self.all_text) - 1]["type"] = "页脚"
            except Exception:
                print(page.page_number)
        else:
            try:
                first_text = str(self.all_text[self.last_num + 2]["inside"])
                end_text = str(self.all_text[len(self.all_text) - 1]["inside"])
                if re.search(first_re, first_text) and "[" not in end_text:
                    self.all_text[self.last_num + 2]["type"] = "页眉"
                if re.search(end_re, end_text) and "[" not in end_text:
                    self.all_text[len(self.all_text) - 1]["type"] = "页脚"
            except Exception:
                print(page.page_number)

        self.last_num = len(self.all_text) - 1

    def pdf_to_json(self):
        """Process pdf."""
        for i in range(len(self.pdf.pages)):
            self.extract_text_and_tables(self.pdf.pages[i])
            logger.info(f"{self.filepath} page {i} extract text success")

    def save_all_text(self, path):
        """Save all text."""
        directory = os.path.dirname(path)
        if not os.path.exists(directory):
            os.makedirs(directory)
        for key in self.all_text.keys():
            with open(path, "a+", encoding="utf-8") as file:
                file.write(json.dumps(self.all_text[key], ensure_ascii=False) + "\n")
